from time import time
from naeural_core.business.base import CVPluginExecutor as BasePlugin

_CONFIG = {
  **BasePlugin.CONFIG,
  
  
  'RESTART_ALERTERS_ON_CONFIG': False,
  'RESTART_SHMEM_ON_CONFIG'   : False,
  'ALLOW_EMPTY_INPUTS'        : True,
  
  'ALLOW_ORIGINAL_IMAGE'      : False,
  
  
  'DEBUG_REST'                : False,
    
  'REQUEST' : None,
  
  'INFO_MANIFEST' : {
    'REQUEST' : {
      'DATA' : {
          'PARAM1' : 'Set value of params',
        },
      'TIMESTAMP' : 'Float value of timestamp',
    }
  },

  'VALIDATION_RULES' : {
    **BasePlugin.CONFIG['VALIDATION_RULES'],

  },  

}

__VER__ = '0.1.0'

class SimpleRestExecutor(BasePlugin):
  CONFIG = _CONFIG
  
  def __init__(self, **kwargs):
    self.__loop_counter = 0
    self.__request_counter = 0
    self.__last_request_timestamp = None
    self.__last_request = None
    self.__time_of_last_request = time()
    self.__time_of_last_manifest = time()
    self.__request_served = False
    self.__first_run_executed = False
    
    super(SimpleRestExecutor, self).__init__(**kwargs)
    return

  
  @property
  def request_count(self):
    return self.__request_counter
  
  def _get_request(self):
    request = self.cfg_request
    data = None    
    _timestamp = None
    if request is not None:
      if isinstance(request, dict) and len(request) > 0:
        # timestamp can be any kind of "nonce" message not necesarelly a actual timestamp
        _timestamp = request.get('TIMESTAMP')
        # we dont actually need repeated_timestamp
        repeated_timestamp = _timestamp is not None and _timestamp == self.__last_request_timestamp
        repeated_request = self.__last_request == request
        maybe_data = request.get('DATA')
        
        if not repeated_request:
          if maybe_data is None:
            self.P("Received request with no 'DATA': {}".format(request), color='r')          
          data = request if maybe_data is None else maybe_data
      else:
        # non dict request ... pass it forward as it is
        if self.__last_request != request:
          self.P("Received non-dict request: {}".format(request), color='r')
          data = request
      # dict or non-dict request
    if data is not None:
      self.__request_counter += 1
      self.__time_of_last_request = time()
      self.__time_of_last_manifest = time()      
      self.__request_served = True
      self.__last_request = request
      self.__last_request_timestamp = _timestamp
    return data
  
  def _on_request(self, data):
    raise NotImplementedError()
    
  def _maybe_cleanup_instance(self):
    ### cleanup any residues from previous runs !
    # clean default image - as it may remain from a previous run and will be continuously delivered
    self.default_image = None
    return
  
  def _needs_manifest(self):
    last_req_time = time() - self.__time_of_last_request
    last_manifest_time = time() - self.__time_of_last_manifest
    result = False
    if self.__request_served:
      if last_req_time > (self.cfg_send_manifest_each * 3):
        result = True
    else:
      if last_manifest_time > self.cfg_send_manifest_each:
        result = True
    return result
  
  
  def _maybe_cancel_original_image(self):
    if self.cfg_add_original_image and not self.cfg_allow_original_image:
      self.P("Found ADD_ORIGINAL_IMAGE=True but ALLOW_ORIGINAL_IMAGE=False. Setting ADD_ORIGINAL_IMAGE=False", color='r')
      self.config_data['ADD_ORIGINAL_IMAGE'] = False
      self.P("  ADD_ORIGINAL_IMAGE={}".format(self.cfg_add_original_image))
    #endif
    return
  
  def _process(self):
    self.__loop_counter += 1
    data = self._get_request()
    payload = None
    self._maybe_cancel_original_image()
    if data is not None:
      self.P("**** Request <{}...> at iter {}".format(str(data)[:50], self.__loop_counter), color='d')
      if self.cfg_debug_rest:
        has_img = self.dataapi_image() is not None
        meta = self.dataapi_all_metadata()
        freq = self.actual_plugin_resolution
        qlen = self.input_queue_size
        ale = self.cfg_allow_empty_inputs
        rwi = self.cfg_run_without_image
        self.P("DEBUG-REST: Image avl: {}, p.freq.: {}, q.len: {}, alw.emt: {}, r.wo.i.: {}\nMetadata: {}".format(
          has_img, freq, qlen, ale, rwi,
          meta
          )
        )
      # endif "DEBUG_REST" : true
      try:
        payload = self._on_request(data)
      except Exception as exc:
        self.P("Exception in on-reques: {}".format(exc), color='r')
        raise exc
      finally:
        self.archive_config_keys(keys=['REQUEST'])

      
      if self.cfg_debug_rest:
        if payload is not None:
          im = vars(payload).get('IMG', [])
          has_img = len(im) > 0
          self.P("DEBUG-REST: Payload IMG: {}, default IMG: {}".format(
            im, self.default_image is not None)
          )
        else:
          self.P("DEBUG-REST: on_request payload None", color='r')
      if payload is not None:
        vars(payload)['TIMESTAMP'] = self.__last_request_timestamp
    elif not self.__first_run_executed or self._needs_manifest():
      __elapsed = min(time() - self.__time_of_last_request, time() - self.__time_of_last_manifest)
      self.__time_of_last_manifest = time()
      self.__request_served = False
      # no need for cleanup as certainly it was already run
      payload = self._create_payload(
        info_manifest=self.cfg_info_manifest
      )
      if self.cfg_log_manifest_send:
        self.P("**** Sending manifest after {:.1f}s/{:.1f}s from last activity".format(
          __elapsed, self.cfg_send_manifest_each), color='d'
        )
    else:
      self._maybe_cleanup_instance()

    if not self.__first_run_executed:
      self.P("* * * * REST-like service up and running * * * * ", color='g')
      self.__first_run_executed = True
    return payload
  
  
  


# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Train a YOLOv5 model on a custom dataset

Usage:
    $ python path/to/train.py --data coco128.yaml --weights yolov5s.pt --img 640
"""

import argparse
import logging
import math
import os
import random
import sys
import time
from copy import deepcopy
from pathlib import Path

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
import yaml
from torch.cuda import amp
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.optim import Adam, SGD, lr_scheduler
from tqdm import tqdm

FILE = Path(__file__).resolve()
ROOT = FILE.parents[0]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
ROOT = ROOT.relative_to(Path.cwd())  # relative

import val  # for end-of-epoch mAP
from models.experimental import attempt_load
from models.yolo import Model
from utils.autoanchor import check_anchors
from utils.datasets import create_dataloader
from utils.general import labels_to_class_weights, increment_path, labels_to_image_weights, init_seeds, \
    strip_optimizer, get_latest_run, check_dataset, check_git_status, check_img_size, check_requirements, \
    check_file, check_yaml, check_suffix, print_args, print_mutation, set_logging, one_cycle, colorstr, methods
from utils.downloads import attempt_download
from utils.loss import ComputeLoss
from utils.plots import plot_labels, plot_evolve
from utils.torch_utils import EarlyStopping, ModelEMA, de_parallel, intersect_dicts, select_device, \
    torch_distributed_zero_first
from utils.loggers.wandb.wandb_utils import check_wandb_resume
from utils.metrics import fitness
from utils.loggers import Loggers
from utils.callbacks import Callbacks

LOGGER = logging.getLogger(__name__)
LOCAL_RANK = int(os.getenv('LOCAL_RANK', -1))  # https://pytorch.org/docs/stable/elastic/run.html
RANK = int(os.getenv('RANK', -1))
WORLD_SIZE = int(os.getenv('WORLD_SIZE', 1))


def train(hyp,  # path/to/hyp.yaml or hyp dictionary
          opt,
          device,
          callbacks
          ):
    save_dir, epochs, batch_size, weights, single_cls, evolve, data, cfg, resume, noval, nosave, workers, freeze, = \
        Path(opt.save_dir), opt.epochs, opt.batch_size, opt.weights, opt.single_cls, opt.evolve, opt.data, opt.cfg, \
        opt.resume, opt.noval, opt.nosave, opt.workers, opt.freeze

    # Directories
    w = save_dir / 'weights'  # weights dir
    (w.parent if evolve else w).mkdir(parents=True, exist_ok=True)  # make dir
    last, best = w / 'last.pt', w / 'best.pt'

    # Hyperparameters
    if isinstance(hyp, str):
        with open(hyp) as f:
            hyp = yaml.safe_load(f)  # load hyps dict
    LOGGER.info(colorstr('hyperparameters: ') + ', '.join(f'{k}={v}' for k, v in hyp.items()))

    # Save run settings
    with open(save_dir / 'hyp.yaml', 'w') as f:
        yaml.safe_dump(hyp, f, sort_keys=False)
    with open(save_dir / 'opt.yaml', 'w') as f:
        yaml.safe_dump(vars(opt), f, sort_keys=False)
    data_dict = None

    # Loggers
    if RANK in [-1, 0]:
        loggers = Loggers(save_dir, weights, opt, hyp, LOGGER)  # loggers instance
        if loggers.wandb:
            data_dict = loggers.wandb.data_dict
            if resume:
                weights, epochs, hyp = opt.weights, opt.epochs, opt.hyp

        # Register actions
        for k in methods(loggers):
            callbacks.register_action(k, callback=getattr(loggers, k))

    # Config
    plots = not evolve  # create plots
    cuda = device.type != 'cpu'
    init_seeds(1 + RANK)
    with torch_distributed_zero_first(RANK):
        data_dict = data_dict or check_dataset(data)  # check if None
    train_path, val_path = data_dict['train'], data_dict['val']
    nc = 1 if single_cls else int(data_dict['nc'])  # number of classes
    names = ['item'] if single_cls and len(data_dict['names']) != 1 else data_dict['names']  # class names
    assert len(names) == nc, f'{len(names)} names found for nc={nc} dataset in {data}'  # check
    is_coco = data.endswith('coco.yaml') and nc == 80  # COCO dataset

    # Model
    check_suffix(weights, '.pt')  # check weights
    pretrained = weights.endswith('.pt')
    if pretrained:
        with torch_distributed_zero_first(RANK):
            weights = attempt_download(weights)  # download if not found locally
        ckpt = torch.load(weights, map_location=device)  # load checkpoint
        model = Model(cfg or ckpt['model'].yaml, ch=3, nc=nc, anchors=hyp.get('anchors')).to(device)  # create
        exclude = ['anchor'] if (cfg or hyp.get('anchors')) and not resume else []  # exclude keys
        csd = ckpt['model'].float().state_dict()  # checkpoint state_dict as FP32
        csd = intersect_dicts(csd, model.state_dict(), exclude=exclude)  # intersect
        model.load_state_dict(csd, strict=False)  # load
        LOGGER.info(f'Transferred {len(csd)}/{len(model.state_dict())} items from {weights}')  # report
    else:
        model = Model(cfg, ch=3, nc=nc, anchors=hyp.get('anchors')).to(device)  # create

    # Freeze
    freeze = [f'model.{x}.' for x in range(freeze)]  # layers to freeze
    for k, v in model.named_parameters():
        v.requires_grad = True  # train all layers
        if any(x in k for x in freeze):
            print(f'freezing {k}')
            v.requires_grad = False

    # Optimizer
    nbs = 64  # nominal batch size
    accumulate = max(round(nbs / batch_size), 1)  # accumulate loss before optimizing
    hyp['weight_decay'] *= batch_size * accumulate / nbs  # scale weight_decay
    LOGGER.info(f"Scaled weight_decay = {hyp['weight_decay']}")

    g0, g1, g2 = [], [], []  # optimizer parameter groups
    for v in model.modules():
        if hasattr(v, 'bias') and isinstance(v.bias, nn.Parameter):  # bias
            g2.append(v.bias)
        if isinstance(v, nn.BatchNorm2d):  # weight (no decay)
            g0.append(v.weight)
        elif hasattr(v, 'weight') and isinstance(v.weight, nn.Parameter):  # weight (with decay)
            g1.append(v.weight)

    if opt.adam:
        optimizer = Adam(g0, lr=hyp['lr0'], betas=(hyp['momentum'], 0.999))  # adjust beta1 to momentum
    else:
        optimizer = SGD(g0, lr=hyp['lr0'], momentum=hyp['momentum'], nesterov=True)

    optimizer.add_param_group({'params': g1, 'weight_decay': hyp['weight_decay']})  # add g1 with weight_decay
    optimizer.add_param_group({'params': g2})  # add g2 (biases)
    LOGGER.info(f"{colorstr('optimizer:')} {type(optimizer).__name__} with parameter groups "
                f"{len(g0)} weight, {len(g1)} weight (no decay), {len(g2)} bias")
    del g0, g1, g2

    # Scheduler
    if opt.linear_lr:
        lf = lambda x: (1 - x / (epochs - 1)) * (1.0 - hyp['lrf']) + hyp['lrf']  # linear
    else:
        lf = one_cycle(1, hyp['lrf'], epochs)  # cosine 1->hyp['lrf']
    scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lf)  # plot_lr_scheduler(optimizer, scheduler, epochs)

    # EMA
    ema = ModelEMA(model) if RANK in [-1, 0] else None

    # Resume
    start_epoch, best_fitness = 0, 0.0
    if pretrained:
        # Optimizer
        if ckpt['optimizer'] is not None:
            optimizer.load_state_dict(ckpt['optimizer'])
            best_fitness = ckpt['best_fitness']

        # EMA
        if ema and ckpt.get('ema'):
            ema.ema.load_state_dict(ckpt['ema'].float().state_dict())
            ema.updates = ckpt['updates']

        # Epochs
        start_epoch = ckpt['epoch'] + 1
        if resume:
            assert start_epoch > 0, f'{weights} training to {epochs} epochs is finished, nothing to resume.'
        if epochs < start_epoch:
            LOGGER.info(f"{weights} has been trained for {ckpt['epoch']} epochs. Fine-tuning for {epochs} more epochs.")
            epochs += ckpt['epoch']  # finetune additional epochs

        del ckpt, csd

    # Image sizes
    gs = max(int(model.stride.max()), 32)  # grid size (max stride)
    nl = model.model[-1].nl  # number of detection layers (used for scaling hyp['obj'])
    imgsz = check_img_size(opt.imgsz, gs, floor=gs * 2)  # verify imgsz is gs-multiple

    # DP mode
    if cuda and RANK == -1 and torch.cuda.device_count() > 1:
        logging.warning('DP not recommended, instead use torch.distributed.run for best DDP Multi-GPU results.\n'
                        'See Multi-GPU Tutorial at https://github.com/ultralytics/yolov5/issues/475 to get started.')
        model = torch.nn.DataParallel(model)

    # SyncBatchNorm
    if opt.sync_bn and cuda and RANK != -1:
        model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model).to(device)
        LOGGER.info('Using SyncBatchNorm()')

    # Trainloader
    train_loader, dataset = create_dataloader(train_path, imgsz, batch_size // WORLD_SIZE, gs, single_cls,
                                              hyp=hyp, augment=True, cache=opt.cache, rect=opt.rect, rank=RANK,
                                              workers=workers, image_weights=opt.image_weights, quad=opt.quad,
                                              prefix=colorstr('train: '))
    mlc = int(np.concatenate(dataset.labels, 0)[:, 0].max())  # max label class
    nb = len(train_loader)  # number of batches
    assert mlc < nc, f'Label class {mlc} exceeds nc={nc} in {data}. Possible class labels are 0-{nc - 1}'

    # Process 0
    if RANK in [-1, 0]:
        val_loader = create_dataloader(val_path, imgsz, batch_size // WORLD_SIZE * 2, gs, single_cls,
                                       hyp=hyp, cache=None if noval else opt.cache, rect=True, rank=-1,
                                       workers=workers, pad=0.5,
                                       prefix=colorstr('val: '))[0]

        if not resume:
            labels = np.concatenate(dataset.labels, 0)
            # c = torch.tensor(labels[:, 0])  # classes
            # cf = torch.bincount(c.long(), minlength=nc) + 1.  # frequency
            # model._initialize_biases(cf.to(device))
            if plots:
                plot_labels(labels, names, save_dir)

            # Anchors
            if not opt.noautoanchor:
                check_anchors(dataset, model=model, thr=hyp['anchor_t'], imgsz=imgsz)
            model.half().float()  # pre-reduce anchor precision

        callbacks.run('on_pretrain_routine_end')

    # DDP mode
    if cuda and RANK != -1:
        model = DDP(model, device_ids=[LOCAL_RANK], output_device=LOCAL_RANK)

    # Model parameters
    hyp['box'] *= 3. / nl  # scale to layers
    hyp['cls'] *= nc / 80. * 3. / nl  # scale to classes and layers
    hyp['obj'] *= (imgsz / 640) ** 2 * 3. / nl  # scale to image size and layers
    hyp['label_smoothing'] = opt.label_smoothing
    model.nc = nc  # attach number of classes to model
    model.hyp = hyp  # attach hyperparameters to model
    model.class_weights = labels_to_class_weights(dataset.labels, nc).to(device) * nc  # attach class weights
    model.names = names

    # Start training
    t0 = time.time()
    nw = max(round(hyp['warmup_epochs'] * nb), 1000)  # number of warmup iterations, max(3 epochs, 1k iterations)
    # nw = min(nw, (epochs - start_epoch) / 2 * nb)  # limit warmup to < 1/2 of training
    last_opt_step = -1
    maps = np.zeros(nc)  # mAP per class
    results = (0, 0, 0, 0, 0, 0, 0)  # P, R, mAP@.5, mAP@.5-.95, val_loss(box, obj, cls)
    scheduler.last_epoch = start_epoch - 1  # do not move
    scaler = amp.GradScaler(enabled=cuda)
    stopper = EarlyStopping(patience=opt.patience)
    compute_loss = ComputeLoss(model)  # init loss class
    LOGGER.info(f'Image sizes {imgsz} train, {imgsz} val\n'
                f'Using {train_loader.num_workers} dataloader workers\n'
                f"Logging results to {colorstr('bold', save_dir)}\n"
                f'Starting training for {epochs} epochs...')
    for epoch in range(start_epoch, epochs):  # epoch ------------------------------------------------------------------
        model.train()

        # Update image weights (optional, single-GPU only)
        if opt.image_weights:
            cw = model.class_weights.cpu().numpy() * (1 - maps) ** 2 / nc  # class weights
            iw = labels_to_image_weights(dataset.labels, nc=nc, class_weights=cw)  # image weights
            dataset.indices = random.choices(range(dataset.n), weights=iw, k=dataset.n)  # rand weighted idx

        # Update mosaic border (optional)
        # b = int(random.uniform(0.25 * imgsz, 0.75 * imgsz + gs) // gs * gs)
        # dataset.mosaic_border = [b - imgsz, -b]  # height, width borders

        mloss = torch.zeros(3, device=device)  # mean losses
        if RANK != -1:
            train_loader.sampler.set_epoch(epoch)
        pbar = enumerate(train_loader)
        LOGGER.info(('\n' + '%10s' * 7) % ('Epoch', 'gpu_mem', 'box', 'obj', 'cls', 'labels', 'img_size'))
        if RANK in [-1, 0]:
            pbar = tqdm(pbar, total=nb)  # progress bar
        optimizer.zero_grad()
        for i, (imgs, targets, paths, _) in pbar:  # batch -------------------------------------------------------------
            ni = i + nb * epoch  # number integrated batches (since train start)
            imgs = imgs.to(device, non_blocking=True).float() / 255.0  # uint8 to float32, 0-255 to 0.0-1.0

            # Warmup
            if ni <= nw:
                xi = [0, nw]  # x interp
                # compute_loss.gr = np.interp(ni, xi, [0.0, 1.0])  # iou loss ratio (obj_loss = 1.0 or iou)
                accumulate = max(1, np.interp(ni, xi, [1, nbs / batch_size]).round())
                for j, x in enumerate(optimizer.param_groups):
                    # bias lr falls from 0.1 to lr0, all other lrs rise from 0.0 to lr0
                    x['lr'] = np.interp(ni, xi, [hyp['warmup_bias_lr'] if j == 2 else 0.0, x['initial_lr'] * lf(epoch)])
                    if 'momentum' in x:
                        x['momentum'] = np.interp(ni, xi, [hyp['warmup_momentum'], hyp['momentum']])

            # Multi-scale
            if opt.multi_scale:
                sz = random.randrange(imgsz * 0.5, imgsz * 1.5 + gs) // gs * gs  # size
                sf = sz / max(imgs.shape[2:])  # scale factor
                if sf != 1:
                    ns = [math.ceil(x * sf / gs) * gs for x in imgs.shape[2:]]  # new shape (stretched to gs-multiple)
                    imgs = nn.functional.interpolate(imgs, size=ns, mode='bilinear', align_corners=False)

            # Forward
            with amp.autocast(enabled=cuda):
                pred = model(imgs)  # forward
                loss, loss_items = compute_loss(pred, targets.to(device))  # loss scaled by batch_size
                if RANK != -1:
                    loss *= WORLD_SIZE  # gradient averaged between devices in DDP mode
                if opt.quad:
                    loss *= 4.

            # Backward
            scaler.scale(loss).backward()

            # Optimize
            if ni - last_opt_step >= accumulate:
                scaler.step(optimizer)  # optimizer.step
                scaler.update()
                optimizer.zero_grad()
                if ema:
                    ema.update(model)
                last_opt_step = ni

            # Log
            if RANK in [-1, 0]:
                mloss = (mloss * i + loss_items) / (i + 1)  # update mean losses
                mem = f'{torch.cuda.memory_reserved() / 1E9 if torch.cuda.is_available() else 0:.3g}G'  # (GB)
                pbar.set_description(('%10s' * 2 + '%10.4g' * 5) % (
                    f'{epoch}/{epochs - 1}', mem, *mloss, targets.shape[0], imgs.shape[-1]))
                callbacks.run('on_train_batch_end', ni, model, imgs, targets, paths, plots, opt.sync_bn)
            # end batch ------------------------------------------------------------------------------------------------

        # Scheduler
        lr = [x['lr'] for x in optimizer.param_groups]  # for loggers
        scheduler.step()

        if RANK in [-1, 0]:
            # mAP
            callbacks.run('on_train_epoch_end', epoch=epoch)
            ema.update_attr(model, include=['yaml', 'nc', 'hyp', 'names', 'stride', 'class_weights'])
            final_epoch = (epoch + 1 == epochs) or stopper.possible_stop
            if not noval or final_epoch:  # Calculate mAP
                results, maps, _ = val.run(data_dict,
                                           batch_size=batch_size // WORLD_SIZE * 2,
                                           imgsz=imgsz,
                                           model=ema.ema,
                                           single_cls=single_cls,
                                           dataloader=val_loader,
                                           save_dir=save_dir,
                                           plots=False,
                                           callbacks=callbacks,
                                           compute_loss=compute_loss)

            # Update best mAP
            fi = fitness(np.array(results).reshape(1, -1))  # weighted combination of [P, R, mAP@.5, mAP@.5-.95]
            if fi > best_fitness:
                best_fitness = fi
            log_vals = list(mloss) + list(results) + lr
            callbacks.run('on_fit_epoch_end', log_vals, epoch, best_fitness, fi)

            # Save model
            if (not nosave) or (final_epoch and not evolve):  # if save
                ckpt = {'epoch': epoch,
                        'best_fitness': best_fitness,
                        'model': deepcopy(de_parallel(model)).half(),
                        'ema': deepcopy(ema.ema).half(),
                        'updates': ema.updates,
                        'optimizer': optimizer.state_dict(),
                        'wandb_id': loggers.wandb.wandb_run.id if loggers.wandb else None}

                # Save last, best and delete
                torch.save(ckpt, last)
                if best_fitness == fi:
                    torch.save(ckpt, best)
                del ckpt
                callbacks.run('on_model_save', last, epoch, final_epoch, best_fitness, fi)

            # Stop Single-GPU
            if RANK == -1 and stopper(epoch=epoch, fitness=fi):
                break

            # Stop DDP TODO: known issues shttps://github.com/ultralytics/yolov5/pull/4576
            # stop = stopper(epoch=epoch, fitness=fi)
            # if RANK == 0:
            #    dist.broadcast_object_list([stop], 0)  # broadcast 'stop' to all ranks

        # Stop DPP
        # with torch_distributed_zero_first(RANK):
        # if stop:
        #    break  # must break all DDP ranks

        # end epoch ----------------------------------------------------------------------------------------------------
    # end training -----------------------------------------------------------------------------------------------------
    if RANK in [-1, 0]:
        LOGGER.info(f'\n{epoch - start_epoch + 1} epochs completed in {(time.time() - t0) / 3600:.3f} hours.')
        for f in last, best:
            if f.exists():
                strip_optimizer(f)  # strip optimizers
                if f is best:
                    LOGGER.info(f'\nValidating {f}...')
                    results, _, _ = val.run(data_dict,
                                            batch_size=batch_size // WORLD_SIZE * 2,
                                            imgsz=imgsz,
                                            model=attempt_load(f, device).half(),
                                            iou_thres=0.65 if is_coco else 0.60,  # best pycocotools results at 0.65
                                            single_cls=single_cls,
                                            dataloader=val_loader,
                                            save_dir=save_dir,
                                            save_json=is_coco,
                                            verbose=True,
                                            plots=True,
                                            callbacks=callbacks,
                                            compute_loss=compute_loss)  # val best model with plots

        callbacks.run('on_train_end', last, best, plots, epoch)
        LOGGER.info(f"Results saved to {colorstr('bold', save_dir)}")

    torch.cuda.empty_cache()
    return results


def parse_opt(known=False):
    parser = argparse.ArgumentParser()
    parser.add_argument('--weights', type=str, default=ROOT / 'yolov5s.pt', help='initial weights path')
    parser.add_argument('--cfg', type=str, default='', help='model.yaml path')
    parser.add_argument('--data', type=str, default=ROOT / 'data/coco128.yaml', help='dataset.yaml path')
    parser.add_argument('--hyp', type=str, default=ROOT / 'data/hyps/hyp.scratch.yaml', help='hyperparameters path')
    parser.add_argument('--epochs', type=int, default=300)
    parser.add_argument('--batch-size', type=int, default=16, help='total batch size for all GPUs')
    parser.add_argument('--imgsz', '--img', '--img-size', type=int, default=640, help='train, val image size (pixels)')
    parser.add_argument('--rect', action='store_true', help='rectangular training')
    parser.add_argument('--resume', nargs='?', const=True, default=False, help='resume most recent training')
    parser.add_argument('--nosave', action='store_true', help='only save final checkpoint')
    parser.add_argument('--noval', action='store_true', help='only validate final epoch')
    parser.add_argument('--noautoanchor', action='store_true', help='disable autoanchor check')
    parser.add_argument('--evolve', type=int, nargs='?', const=300, help='evolve hyperparameters for x generations')
    parser.add_argument('--bucket', type=str, default='', help='gsutil bucket')
    parser.add_argument('--cache', type=str, nargs='?', const='ram', help='--cache images in "ram" (default) or "disk"')
    parser.add_argument('--image-weights', action='store_true', help='use weighted image selection for training')
    parser.add_argument('--device', default='', help='cuda device, i.e. 0 or 0,1,2,3 or cpu')
    parser.add_argument('--multi-scale', action='store_true', help='vary img-size +/- 50%%')
    parser.add_argument('--single-cls', action='store_true', help='train multi-class data as single-class')
    parser.add_argument('--adam', action='store_true', help='use torch.optim.Adam() optimizer')
    parser.add_argument('--sync-bn', action='store_true', help='use SyncBatchNorm, only available in DDP mode')
    parser.add_argument('--workers', type=int, default=8, help='maximum number of dataloader workers')
    parser.add_argument('--entity', default=None, help='W&B entity')
    parser.add_argument('--project', default=ROOT / 'runs/train', help='save to project/name')
    parser.add_argument('--name', default='exp', help='save to project/name')
    parser.add_argument('--exist-ok', action='store_true', help='existing project/name ok, do not increment')
    parser.add_argument('--quad', action='store_true', help='quad dataloader')
    parser.add_argument('--linear-lr', action='store_true', help='linear LR')
    parser.add_argument('--label-smoothing', type=float, default=0.0, help='Label smoothing epsilon')
    parser.add_argument('--upload_dataset', action='store_true', help='Upload dataset as W&B artifact table')
    parser.add_argument('--bbox_interval', type=int, default=-1, help='Set bounding-box image logging interval for W&B')
    parser.add_argument('--save_period', type=int, default=-1, help='Log model after every "save_period" epoch')
    parser.add_argument('--artifact_alias', type=str, default="latest", help='version of dataset artifact to be used')
    parser.add_argument('--local_rank', type=int, default=-1, help='DDP parameter, do not modify')
    parser.add_argument('--freeze', type=int, default=0, help='Number of layers to freeze. backbone=10, all=24')
    parser.add_argument('--patience', type=int, default=100, help='EarlyStopping patience (epochs without improvement)')
    opt = parser.parse_known_args()[0] if known else parser.parse_args()
    return opt


def main(opt, callbacks=Callbacks()):
    # Checks
    set_logging(RANK)
    if RANK in [-1, 0]:
        print_args(FILE.stem, opt)
        check_git_status()
        check_requirements(exclude=['thop'])

    # Resume
    if opt.resume and not check_wandb_resume(opt) and not opt.evolve:  # resume an interrupted run
        ckpt = opt.resume if isinstance(opt.resume, str) else get_latest_run()  # specified or most recent path
        assert os.path.isfile(ckpt), 'ERROR: --resume checkpoint does not exist'
        with open(Path(ckpt).parent.parent / 'opt.yaml') as f:
            opt = argparse.Namespace(**yaml.safe_load(f))  # replace
        opt.cfg, opt.weights, opt.resume = '', ckpt, True  # reinstate
        LOGGER.info(f'Resuming training from {ckpt}')
    else:
        opt.data, opt.cfg, opt.hyp, opt.weights, opt.project = \
            check_file(opt.data), check_yaml(opt.cfg), check_yaml(opt.hyp), str(opt.weights), str(opt.project)  # checks
        assert len(opt.cfg) or len(opt.weights), 'either --cfg or --weights must be specified'
        if opt.evolve:
            opt.project = str(ROOT / 'runs/evolve')
            opt.exist_ok, opt.resume = opt.resume, False  # pass resume to exist_ok and disable resume
        opt.save_dir = str(increment_path(Path(opt.project) / opt.name, exist_ok=opt.exist_ok))

    # DDP mode
    device = select_device(opt.device, batch_size=opt.batch_size)
    if LOCAL_RANK != -1:
        from datetime import timedelta
        assert torch.cuda.device_count() > LOCAL_RANK, 'insufficient CUDA devices for DDP command'
        assert opt.batch_size % WORLD_SIZE == 0, '--batch-size must be multiple of CUDA device count'
        assert not opt.image_weights, '--image-weights argument is not compatible with DDP training'
        assert not opt.evolve, '--evolve argument is not compatible with DDP training'
        torch.cuda.set_device(LOCAL_RANK)
        device = torch.device('cuda', LOCAL_RANK)
        dist.init_process_group(backend="nccl" if dist.is_nccl_available() else "gloo")

    # Train
    if not opt.evolve:
        train(opt.hyp, opt, device, callbacks)
        if WORLD_SIZE > 1 and RANK == 0:
            LOGGER.info('Destroying process group... ')
            dist.destroy_process_group()

    # Evolve hyperparameters (optional)
    else:
        # Hyperparameter evolution metadata (mutation scale 0-1, lower_limit, upper_limit)
        meta = {'lr0': (1, 1e-5, 1e-1),  # initial learning rate (SGD=1E-2, Adam=1E-3)
                'lrf': (1, 0.01, 1.0),  # final OneCycleLR learning rate (lr0 * lrf)
                'momentum': (0.3, 0.6, 0.98),  # SGD momentum/Adam beta1
                'weight_decay': (1, 0.0, 0.001),  # optimizer weight decay
                'warmup_epochs': (1, 0.0, 5.0),  # warmup epochs (fractions ok)
                'warmup_momentum': (1, 0.0, 0.95),  # warmup initial momentum
                'warmup_bias_lr': (1, 0.0, 0.2),  # warmup initial bias lr
                'box': (1, 0.02, 0.2),  # box loss gain
                'cls': (1, 0.2, 4.0),  # cls loss gain
                'cls_pw': (1, 0.5, 2.0),  # cls BCELoss positive_weight
                'obj': (1, 0.2, 4.0),  # obj loss gain (scale with pixels)
                'obj_pw': (1, 0.5, 2.0),  # obj BCELoss positive_weight
                'iou_t': (0, 0.1, 0.7),  # IoU training threshold
                'anchor_t': (1, 2.0, 8.0),  # anchor-multiple threshold
                'anchors': (2, 2.0, 10.0),  # anchors per output grid (0 to ignore)
                'fl_gamma': (0, 0.0, 2.0),  # focal loss gamma (efficientDet default gamma=1.5)
                'hsv_h': (1, 0.0, 0.1),  # image HSV-Hue augmentation (fraction)
                'hsv_s': (1, 0.0, 0.9),  # image HSV-Saturation augmentation (fraction)
                'hsv_v': (1, 0.0, 0.9),  # image HSV-Value augmentation (fraction)
                'degrees': (1, 0.0, 45.0),  # image rotation (+/- deg)
                'translate': (1, 0.0, 0.9),  # image translation (+/- fraction)
                'scale': (1, 0.0, 0.9),  # image scale (+/- gain)
                'shear': (1, 0.0, 10.0),  # image shear (+/- deg)
                'perspective': (0, 0.0, 0.001),  # image perspective (+/- fraction), range 0-0.001
                'flipud': (1, 0.0, 1.0),  # image flip up-down (probability)
                'fliplr': (0, 0.0, 1.0),  # image flip left-right (probability)
                'mosaic': (1, 0.0, 1.0),  # image mixup (probability)
                'mixup': (1, 0.0, 1.0),  # image mixup (probability)
                'copy_paste': (1, 0.0, 1.0)}  # segment copy-paste (probability)

        with open(opt.hyp) as f:
            hyp = yaml.safe_load(f)  # load hyps dict
            if 'anchors' not in hyp:  # anchors commented in hyp.yaml
                hyp['anchors'] = 3
        opt.noval, opt.nosave, save_dir = True, True, Path(opt.save_dir)  # only val/save final epoch
        # ei = [isinstance(x, (int, float)) for x in hyp.values()]  # evolvable indices
        evolve_yaml, evolve_csv = save_dir / 'hyp_evolve.yaml', save_dir / 'evolve.csv'
        if opt.bucket:
            os.system(f'gsutil cp gs://{opt.bucket}/evolve.csv {save_dir}')  # download evolve.csv if exists

        for _ in range(opt.evolve):  # generations to evolve
            if evolve_csv.exists():  # if evolve.csv exists: select best hyps and mutate
                # Select parent(s)
                parent = 'single'  # parent selection method: 'single' or 'weighted'
                x = np.loadtxt(evolve_csv, ndmin=2, delimiter=',', skiprows=1)
                n = min(5, len(x))  # number of previous results to consider
                x = x[np.argsort(-fitness(x))][:n]  # top n mutations
                w = fitness(x) - fitness(x).min() + 1E-6  # weights (sum > 0)
                if parent == 'single' or len(x) == 1:
                    # x = x[random.randint(0, n - 1)]  # random selection
                    x = x[random.choices(range(n), weights=w)[0]]  # weighted selection
                elif parent == 'weighted':
                    x = (x * w.reshape(n, 1)).sum(0) / w.sum()  # weighted combination

                # Mutate
                mp, s = 0.8, 0.2  # mutation probability, sigma
                npr = np.random
                npr.seed(int(time.time()))
                g = np.array([meta[k][0] for k in hyp.keys()])  # gains 0-1
                ng = len(meta)
                v = np.ones(ng)
                while all(v == 1):  # mutate until a change occurs (prevent duplicates)
                    v = (g * (npr.random(ng) < mp) * npr.randn(ng) * npr.random() * s + 1).clip(0.3, 3.0)
                for i, k in enumerate(hyp.keys()):  # plt.hist(v.ravel(), 300)
                    hyp[k] = float(x[i + 7] * v[i])  # mutate

            # Constrain to limits
            for k, v in meta.items():
                hyp[k] = max(hyp[k], v[1])  # lower limit
                hyp[k] = min(hyp[k], v[2])  # upper limit
                hyp[k] = round(hyp[k], 5)  # significant digits

            # Train mutation
            results = train(hyp.copy(), opt, device, callbacks)

            # Write mutation results
            print_mutation(results, hyp.copy(), save_dir, opt.bucket)

        # Plot results
        plot_evolve(evolve_csv)
        print(f'Hyperparameter evolution finished\n'
              f"Results saved to {colorstr('bold', save_dir)}\n"
              f'Use best hyperparameters example: $ python train.py --hyp {evolve_yaml}')


def run(**kwargs):
    # Usage: import train; train.run(data='coco128.yaml', imgsz=320, weights='yolov5m.pt')
    opt = parse_opt(True)
    for k, v in kwargs.items():
        setattr(opt, k, v)
    main(opt)


if __name__ == "__main__":
    opt = parse_opt()
    main(opt)



from naeural_core.data.base import BaseStructuredDataCapture


_CONFIG = {
  **BaseStructuredDataCapture.CONFIG,
  
  'DESCRIPTION' : 'This pipeline generates a simple artificial counter at each capture iteration ',
  
  'VALIDATION_RULES' : {
    **BaseStructuredDataCapture.CONFIG['VALIDATION_RULES'],
  },
}

class CounterStreamDataCapture(BaseStructuredDataCapture):
  CONFIG = _CONFIG
      
  def data_step(self):
    data = {'counter' : self._metadata.cap_struct_data_count}
    return data


"""
{
    "NAME": "CroppedUsbCameraMeta",
    "DESCRIPTION": "Test for croppers and other stuff",
    "TYPE": "SingleCropMetaStream",
    
    "COLLECTED_STREAMS" : ["UsbCamera"],
    
    "CROP_TOP"         : 200,
    "CROP_LEFT"        : 300,
    "CROP_BOTTOM"      : 400,
    "CROP_RIGHT"       : 639,    


    "PLUGINS": [
        {
            "INSTANCES": [
                {
                    "INSTANCE_ID": "WITNES_ON_META",
                    "PROCESS_DELAY": 1
                }
            ],
            "SIGNATURE": "VIEW_SCENE_01"
        }
    ]
} 

"""

from naeural_core.data.default.meta_stream import MetaStreamDataCapture as BaseClass

_CONFIG = {
  **BaseClass.CONFIG,  
  
  'CROP_TOP'         : 0,
  'CROP_LEFT'        : 0,
  'CROP_BOTTOM'      : 100,
  'CROP_RIGHT'       : 100,
  
  'VALIDATION_RULES' : {
    **BaseClass.CONFIG['VALIDATION_RULES'],
    
  },
}

class SingleCropMetaStreamDataCapture(BaseClass):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(SingleCropMetaStreamDataCapture, self).__init__(**kwargs)
    return
  
  def validate_coords(self):
    delta = 50
    if (self.cfg_crop_top + delta) >= self.cfg_crop_bottom:
      self.add_error("CROP_TOP:{} BIGGER THAN CROP_BOTTOM:{}+{}".format(
        self.cfg_crop_top, self.cfg_crop_bottom, delta)
      )            
    if (self.cfg_crop_left + delta) >= self.cfg_crop_right:
      self.add_error("CROP_LEFT:{} BIGGER THAN CROP_RIGHT:{}+{}".format(
        self.cfg_crop_left, self.cfg_crop_right, delta)
      )
    return
  
  
  def post_process_inputs(self, lst_stream_inputs):
    if len(lst_stream_inputs) > 0:
      # work on just one image
      img_orig = lst_stream_inputs[0]['IMG']
      _t = self.cfg_crop_top
      _l = self.cfg_crop_left
      _r = self.cfg_crop_right
      _b = self.cfg_crop_bottom
      if _b > img_orig.shape[0] or _r > img_orig.shape[1]:
        raise ValueError("The proposed crop bottom/right coords of {} are outside of the given video stream resolution of {}".format(
          (_b, _r), img_orig.shape)
        )
      img_crop = img_orig[_t:_b, _l:_r]
      lst_stream_inputs[0]['IMG'] = img_crop
      lst_stream_inputs[0]['METADATA']['original_image'] = img_orig
      lst_stream_inputs[0]['METADATA']['offset_h'] = _t
      lst_stream_inputs[0]['METADATA']['offset_w'] = _l
      
    # end if min 1 image
    return lst_stream_inputs
      
    
  





import cv2
import numpy as np
import matplotlib.pyplot as plt
import math

rotation_angle = -math.pi / 3
sh = -0.2
sv = -0.8

translation_mat = np.array([[1, 0, -1920/2], [0, 1, -1080/2], [0, 0, 1]])
rotation_mat = np.array([[math.cos(rotation_angle), math.sin(rotation_angle), 0], [-math.sin(rotation_angle), math.cos(rotation_angle), 0], [0, 0, 1]])
shear_h_mat = np.array([[1, sh, 0], [0, 1, 0], [0, 0, 1]])
shear_v_mat = np.array([[1, 0, 0], [sv, 1, 0], [0, 0, 1]])
translation_inv_mat = np.array([[1, 0, 1920/2], [0, 1, 1080/2], [0, 0, 1]])
tl = np.array([0,0,1])
tr = [1,-1,1]
bl = [-1, 1, 1]
br = [1,1, 1]

matrix = translation_inv_mat @ shear_v_mat @ shear_h_mat @ rotation_mat @ translation_mat

img = cv2.imread('/home/work/Pictures/vlcsnap-2021-07-02-15h29m06s352.png')
img_transformed = cv2.warpPerspective(img,matrix,(1920,1080))
plt.imshow(img_transformed[:, :, ::-1])  # Show results
plt.show()


import sys
import matplotlib.pyplot as plt
import time
import numpy as np
from scipy.spatial import ConvexHull

eps = np.random.rand(1)/100


def on_segment(p, q, r):
    return ((q[0] <= max(p[0], r[0])) and (q[0] >= min(p[0], r[0]))
            and q[1] <= max(p[1] , r[1]) and q[1] >= min(p[1], r[1]))

def orientation(p, q, r):
    val = (q[1] - p[1]) * (r[0] - q[0]) - \
          (q[0] - p[0]) * (r[1] - q[1])

    if val == 0:
        return 0
    elif val > 0:
        return 1
    else:
        return -1

def intersect(p1, q1, p2, q2):
    o1 = orientation(p1, q1, p2)
    o2 = orientation(p1, q1, q2)
    o3 = orientation(p2, q2, p1)
    o4 = orientation(p2, q2, q1)

    if (o1 != o2 and o3 != o4):
        return True
    
    if (o1 == 0 and on_segment(p1, p2, q1)):
        return True
    if (o2 == 0 and on_segment(p1, q2, q1)):
        return True
    if (o3 == 0 and on_segment(p2, p1, q2)):
        return True
    if (o4 == 0 and on_segment(p2, q1, q2)):
        return True

    return False

def is_inside(pts, p):
    # taken from https://www.geeksforgeeks.org/how-to-check-if-a-given-point-lies-inside-a-polygon/

    eps = 0.0001
    # to avoid a special case where ray from point to infinity crosses the two vertices of the polygon.
    # and all these three points become collinear.
    p[0] += eps
    p[1] += eps

    n = len(pts)
    extreme = (sys.maxsize, p[1])

    count = 0
    i = 0

    for i in range(n):
        _next = (i + 1) % n

        if intersect(pts[i], pts[_next], p, extreme):
            if orientation(pts[i], p, pts[_next]) == 0:
                return on_segment(pts[i], p, pts[_next])

            count += 1

    return bool(count & 1)

def get_line(pt1, pt2):
    if pt1[0] == pt2[0]:
        return sys.maxsize, pt1[0]
    
    m = (pt2[1] - pt1[1])/(pt2[0] - pt1[0])
    c = (pt1[1]*pt2[0] - pt2[1]*pt1[0]) / (pt2[0] - pt1[0])

    return m, c

def get_line_intersection(m1, c1, m2, c2):
    x = (c2 - c1)/(m1 - m2)
    y = m1 * x + c1

    return x, y

def get_length(pt1, pt2):
    a = pt1[0] - pt2[0]
    a = np.square(a)
    b = pt1[1] - pt2[1]
    b = np.square(b)

    return np.sqrt(a + b)

def is_on_segment(pt1, pt2, pt3, pt4, pt, eps=-0.001):
    x1, y1 = pt1
    x2, y2 = pt2
    x3, y3 = pt3
    x4, y4 = pt4
    x, y = pt
    return eps<=max(y1, y2)-y and y-min(y1, y2)>=eps and eps<=max(y3, y4)-y and y-min(y3, y4)>=eps and \
        eps<=max(x1, x2)-x and x-min(x1, x2)>=eps and eps<=max(x3, x4)-x and x-min(x3, x4)>=eps

def get_intersection(poly1, poly2):
    # print("\n\n")
    # print(poly1)
    # print(poly2)

    intersecting_coords = []
    for c in poly1[:-1]:
        if is_inside(poly2[:-1], c):
            intersecting_coords.append((c[0], c[1]))
            plt.scatter(c[0], c[1])   
            
    for c in poly2[:-1]:
        if is_inside(poly1[:-1], c):
            intersecting_coords.append((c[0], c[1]))
            plt.scatter(c[0], c[1])
            
    for i1, c1 in enumerate(poly1[:-1]):
        for i2, c2 in enumerate(poly2[:-1]):
            pt1, pt2 = (poly1[i1], poly1[(i1+1)%len(poly1)])
            x1, y1 = pt1
            x2, y2 = pt2
            pt3, pt4 = (poly2[i2], poly2[(i2+1)%len(poly2)])
            x3, y3 = pt3
            x4, y4 = pt4
            m1, c1 = get_line(pt1, pt2)
            m2, c2 = get_line(pt3, pt4)

            if m1 == m2 and c1 == c2:
                if get_length(pt1, pt2) < get_length(pt3, pt4):
                    intersecting_coords.append(pt1)
                    intersecting_coords.append(pt2)
                    plt.scatter(pt1, pt2)
                else:
                    intersecting_coords.append(pt3)
                    intersecting_coords.append(pt4)
                    plt.scatter(pt3, pt4)
            elif m1!=m2:

                if m1 == sys.maxsize:
                    x = x1
                    y = m2 * x1 + c2
                    if is_on_segment(pt1, pt2, pt3, pt4, (x, y)):
                        intersecting_coords.append((x, y))
                        plt.scatter(x, y)
                elif m2 == sys.maxsize:
                    x = x3
                    y = m1 * x3 + c1
                    if is_on_segment(pt1, pt2, pt3, pt4, (x, y)):
                        intersecting_coords.append((x, y))
                        plt.scatter(x, y)
                elif m1 == 0:
                    y = y1
                    x = (y1 - c2) / m2
                    if is_on_segment(pt1, pt2, pt3, pt4, (x, y)):
                        intersecting_coords.append((x, y))
                        plt.scatter(x, y)
                elif m2 == 0:
                    y = y3
                    x = (y3 - c1) / m1
                    if is_on_segment(pt1, pt2, pt3, pt4, (x, y)):
                        intersecting_coords.append((x, y))
                        plt.scatter(x, y)
                else:
                    x, y = get_line_intersection(m1, c1, m2, c2) 

                    _m1 = y - pt1[1]
                    _m1 /= (x - pt1[0])
                    _m2 = y - pt2[1]
                    _m2 /= (x - pt2[0])

                    _m3 = y - pt3[1]
                    _m3 /= (x - pt3[0])
                    _m4 = y - pt4[1]
                    _m4 /= (x - pt4[0])
                    if _m1 * _m2 < 0 and _m3 * _m4 < 0:
                        intersecting_coords.append((x, y))
                        plt.scatter(x, y)
            else:
                print('do not intersect parallel')        

    intersecting_coords = np.array(intersecting_coords)
    hull = ConvexHull(intersecting_coords)
    intersecting_coords = intersecting_coords[hull.vertices]
    intersecting_coords = np.append(intersecting_coords, [intersecting_coords[0]], axis=0)
    x, y = intersecting_coords.T

    print(intersecting_coords)
    print(len(intersecting_coords))
    x1, y1 = poly1.T
    x2, y2 = poly2.T
    plt.plot(x1, y1)
    plt.plot(x2, y2)
    plt.plot(x, y, c='k')
    plt.show()

poly1 = np.array([[451.8107,  173.44763],
 [531.7912,  173.45604],
 [533.50757, 222.04575],
 [469, 210],
 [451.8107,  173.44763]])

poly2 = np.array([[470,  180  ],
 [500, 180  ],
 [500, 220],
 [470,  220],
 [470,  180  ]])
start_time = time.time()
get_intersection(poly1, poly2)
print("--- %s seconds ---" % (time.time() - start_time))


import os, platform, subprocess, re

from naeural_core import Logger

def get_processor_platform():
    if platform.system() == "Windows":
      return platform.processor()
    elif platform.system() == "Darwin":
      os.environ['PATH'] = os.environ['PATH'] + os.pathsep + '/usr/sbin'
      command ="sysctl -n machdep.cpu.brand_string"
      return subprocess.check_output(command).strip()
    elif platform.system() == "Linux":
      command = "cat /proc/cpuinfo"
      all_info = subprocess.check_output(command, shell=True).decode().strip()
      for line in all_info.split("\n"):
        if "model name" in line:
          return re.sub( ".*model name.*:", "", line,1)
    return ""

if __name__ == '__main__':
  print(get_processor_platform())  
  
  l = Logger('PROC', base_folder='.', app_folder='_cache')


import json
import gc
import os

import traceback

from collections import OrderedDict

from time import time, sleep
from naeural_core import constants as ct
from naeural_core import Logger
from naeural_core.manager import Manager

from collections import deque, defaultdict

class BusinessManager(Manager):

  def __init__(self, log : Logger, owner, shmem, environment_variables=None, run_on_threads=True, **kwargs):
    self.shmem = shmem
    self.shmem['get_active_plugins_instances'] = self.get_active_plugins_instances
    self.plugins_shmem = {}
    self.owner = owner
    self.__netmon_instance = None
    self._dct_config_streams = None
    self.is_supervisor_node = self.owner.is_supervisor_node
    self.__evm_network = self.owner.evm_network
    self.shmem['is_supervisor_node'] = self.is_supervisor_node
    self.shmem['__evm_network'] = self.__evm_network
    self.comm_shared_memory = {
      'payloads' : {},
      'commands' : {},
    }
    self._run_on_threads = run_on_threads
    self._environment_variables = environment_variables

    self.dct_serving_processes_startup_params = None
    self.dct_serving_processes_details = None
    
    
    ### each business plugin will be kept having key a hash created based on (stream_name, signature, config_instance)
    ### so whenever a param changes from config_instance we know to deallocate the old business plugin and initialize a new one
    self._dct_current_instances = {}
    self._dct_hash_mappings = {}
    self._dct_stop_timings = {}
    self._dct_instance_hash_log = OrderedDict()


    self._graceful_stop_instances = defaultdict(lambda: 0)
    super(BusinessManager, self).__init__(log=log, prefix_log='[BIZM]', **kwargs)
    return
  
  def _str_to_bool(self, s):
    result = False
    if isinstance(s, bool):
      result = s
    if s is None:
      result = False
    if isinstance(s, int):
      result = bool(s)
    if isinstance(s, str):
      s = s.lower()
      result = s == 'true'
    return result
  

  def startup(self):
    super().startup()
    self._dct_current_instances = self._dct_subalterns # this allows usage of `self.get_subaltern(instance_hash)`
    return

  @property
  def current_instances(self):
    return self._dct_current_instances

  def update_streams(self, dct_config_streams):
    self.owner.set_loop_stage('2.bm.refresh.entry_update_streams')
    self._dct_config_streams = dct_config_streams
    self.owner.set_loop_stage('2.bm.refresh._check_instances')
    current_instances = self._check_instances()
    self.owner.set_loop_stage('2.bm.refresh._deallocate_unused_instances')
    self._deallocate_unused_instances(current_instances)
    self.owner.set_loop_stage('2.bm.refresh.fetch_ai_engines')
    in_use_ai_engines = self.fetch_ai_engines()
    return in_use_ai_engines
    

  def get_active_plugins_instances(self, as_dict=True):
    active = []
    instances = list(self._dct_current_instances.keys())
    for instance in instances:
      plg = self._dct_current_instances.get(instance)
      if plg is None:
        continue
      sid, sign, iid, apr, it, et, lct, fet, let, owh, cei, cpi, lpt, tpc = [None] * 14
      info = None
      
      try:
        # this section MUST be protected as it will call plugin code
        sid = plg._stream_id
        sign = plg._signature
        iid = plg.cfg_instance_id
        
        pdl = plg.cfg_process_delay

        apr = plg.actual_plugin_resolution
        it = plg.init_timestamp
        et = plg.exec_timestamp
        lct = plg.last_config_timestamp
        fet = plg.first_error_time
        let = plg.last_error_time
        owh = plg.is_outside_working_hours # modified within the plugin loop - DO NOT use `plg.outside_working_hours`
        cei = plg.current_exec_iteration
        cpi = plg.current_process_iteration
        lpt = plg.last_payload_time_str
        tpc = plg.total_payload_count
      except Exception as exc:
        info = "Error while retrieving data: {}".format(exc)
      #end try


      if as_dict:
        plg_info = {
          ct.HB.ACTIVE_PLUGINS_INFO.STREAM_ID                  : sid,
          ct.HB.ACTIVE_PLUGINS_INFO.SIGNATURE                  : sign,
          ct.HB.ACTIVE_PLUGINS_INFO.INSTANCE_ID                : iid,
          
          ct.HB.ACTIVE_PLUGINS_INFO.PROCESS_DELAY              : pdl,

          ct.HB.ACTIVE_PLUGINS_INFO.FREQUENCY                  : apr,
          ct.HB.ACTIVE_PLUGINS_INFO.INIT_TIMESTAMP             : it,
          ct.HB.ACTIVE_PLUGINS_INFO.EXEC_TIMESTAMP             : et, # this is the last exec timestamp. It is updated at the end of the exec
          ct.HB.ACTIVE_PLUGINS_INFO.LAST_CONFIG_TIMESTAMP      : lct,
          ct.HB.ACTIVE_PLUGINS_INFO.FIRST_ERROR_TIME           : fet,
          ct.HB.ACTIVE_PLUGINS_INFO.LAST_ERROR_TIME            : let,
          ct.HB.ACTIVE_PLUGINS_INFO.OUTSIDE_WORKING_HOURS      : owh,
          ct.HB.ACTIVE_PLUGINS_INFO.CURRENT_PROCESS_ITERATION  : cpi,
          ct.HB.ACTIVE_PLUGINS_INFO.CURRENT_EXEC_ITERATION     : cei,
          ct.HB.ACTIVE_PLUGINS_INFO.LAST_PAYLOAD_TIME          : lpt,
          ct.HB.ACTIVE_PLUGINS_INFO.TOTAL_PAYLOAD_COUNT        : tpc,

          ct.HB.ACTIVE_PLUGINS_INFO.INFO                       : info,
          }
      else:
        plg_info = (
          sid, sign, iid,
          apr, it, et, lct, fet, let, owh, cpi, cei, lpt, tpc,
          info
        )
      #endif dict or tuple
      active.append(plg_info)
    # end for
    return active


  def get_total_payload_count(self):
    instances = list(self._dct_current_instances.keys())
    total = 0
    for instance in instances:
      plg = self._dct_current_instances.get(instance)
      if plg is None:
        continue
      tpc = plg.total_payload_count
      total += tpc
    return total

  def set_loop_stage(self, s):
    self.owner.set_loop_stage(s)
    return


  def get_business_instance_identification(self, instance_hash):
    return self._dct_hash_mappings.get(instance_hash)

  def get_business_instance_hash(self, stream_name, signature, instance_id):
    dct_inv = {v : k for k,v in self._dct_hash_mappings.items()}
    return dct_inv.get((stream_name, signature, instance_id))

  def get_current_jobs(self):
    current_pipeline_names = list(self._dct_config_streams.keys())
    # now prioritize the "admin_pipeline" (ct.CONST_ADMIN_PIPELINE_NAME) to be the first one
    if ct.CONST_ADMIN_PIPELINE_NAME in current_pipeline_names:
      current_pipeline_names.remove(ct.CONST_ADMIN_PIPELINE_NAME)
      current_pipeline_names.insert(0, ct.CONST_ADMIN_PIPELINE_NAME)   
    #endif prioritize admin pipeline 
    jobs = []
    for pipeline_name in current_pipeline_names:
      pipeline_config = self._dct_config_streams[pipeline_name]

      initiator_addr = pipeline_config.get(ct.CONFIG_STREAM.K_INITIATOR_ADDR, None) 
      initiator_id = pipeline_config.get(ct.CONFIG_STREAM.K_INITIATOR_ID, None)

      modified_by_addr = pipeline_config.get(ct.CONFIG_STREAM.K_MODIFIED_BY_ADDR, None)
      modified_by_id = pipeline_config.get(ct.CONFIG_STREAM.K_MODIFIED_BY_ID, None)

      lst_config_plugins = pipeline_config[ct.CONFIG_STREAM.K_PLUGINS]

      if pipeline_name == ct.CONST_ADMIN_PIPELINE_NAME:
        # Netmon plugin should be the first one in the admin pipeline.
        # This is because the netmon will toggle supervisor node status for every other plugin instance.
        netmon_config_idx, netmon_config = None, None
        for idx, config_plugin in enumerate(lst_config_plugins):
          signature = config_plugin[ct.CONFIG_PLUGIN.K_SIGNATURE]
          if signature == ct.ADMIN_PIPELINE_NETMON:
            netmon_config_idx = idx
            netmon_config = config_plugin
            break
        # endfor plugin_configs
        # Now move the netmon plugin to the first position.
        if netmon_config is not None:
          lst_config_plugins.pop(netmon_config_idx)
          lst_config_plugins.insert(0, netmon_config)
        # endif netmon_config found
      # endif admin pipeline

      session_id = pipeline_config.get(ct.CONFIG_STREAM.K_SESSION_ID, None)
      
      for config_plugin in lst_config_plugins:
        lst_config_instances = config_plugin[ct.CONFIG_PLUGIN.K_INSTANCES]
        signature = config_plugin[ct.CONFIG_PLUGIN.K_SIGNATURE]
        for config_instance in lst_config_instances:
          instance_id = config_instance[ct.CONFIG_INSTANCE.K_INSTANCE_ID]
          jobs.append((initiator_addr, initiator_id, modified_by_addr, modified_by_id, session_id, pipeline_name, signature, instance_id, config_instance))
    return jobs

  def _check_instances(self):
    """
    IMPORTANT: this code section is critical wrt overall main loop functioning!
    """
    current_instances = []
    self.set_loop_stage('2.bm.refresh._check_instances.get_current_jobs')
    all_jobs = self.get_current_jobs()
    for idx_job, (initiator_addr, initiator_id, modified_by_addr, modified_by_id, session_id, stream_name, signature, instance_id, upstream_config) in enumerate(all_jobs):
      obj_identification = (stream_name, signature, instance_id)
      instance_hash = self.log.hash_object(obj_identification, size=5)
      self._dct_hash_mappings[instance_hash] = obj_identification
      current_instances.append(instance_hash)
      if instance_hash not in self._dct_current_instances:
        self._dct_instance_hash_log[instance_hash] = {
          ct.PAYLOAD_DATA.INITIATOR_ID  : initiator_id,
          ct.PAYLOAD_DATA.SESSION_ID    : session_id,
          ct.PAYLOAD_DATA.SIGNATURE     : signature,
          ct.PAYLOAD_DATA.STREAM_NAME   : stream_name,
          ct.PAYLOAD_DATA.INSTANCE_ID   : instance_id,
        }

        self.P(" * * * * Init biz plugin {}:{} * * * *".format(signature, instance_id), color='b')
        self.set_loop_stage('2.bm.refresh.get_class.{}:{}'.format(signature,instance_id))
        if 'update_monitor' in signature.lower():
          print('debug')
        _module_name, _class_name, _cls_def, _config_dict = self._get_module_name_and_class(
          locations=ct.PLUGIN_SEARCH.LOC_BIZ_PLUGINS,
          name=signature,
          suffix=ct.PLUGIN_SEARCH.SUFFIX_BIZ_PLUGINS,
          verbose=0,
          safety_check=True,  # perform safety check on custom biz plugins
          safe_locations=ct.PLUGIN_SEARCH.SAFE_BIZ_PLUGINS,
          safe_imports=ct.PLUGIN_SEARCH.SAFE_BIZ_IMPORTS
        )

        self.set_loop_stage('2.bm.refresh.check_class.{}:{}'.format(signature,instance_id))
        
        if _cls_def is None:
          self._dct_current_instances[instance_hash] = None
          msg = "Error loading business plugin <{}:{}> - No code/script defined.".format(signature, instance_id)
          self.P(msg + " on stream {}".format(stream_name), color='r')
          self._create_notification(
            notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
            msg=msg,
            stream_name=stream_name,
            info="No code/script defined for business plugin '{}' in {} or plugin is invalid (node {})".format(
              signature, ct.PLUGIN_SEARCH.LOC_BIZ_PLUGINS, "is currently SECURED" if self.owner.is_secured else "is currently UNSECURED!"
            )
          )
          continue
        #endif

        self.comm_shared_memory['payloads'][instance_hash] = deque(maxlen=100)
        self.comm_shared_memory['commands'][instance_hash] = deque(maxlen=100)

        try:
          self.set_loop_stage('2.bm.refresh.call_class.{}:{}:{}'.format(stream_name, signature, instance_id))
          self.shmem['__set_loop_stage_func'] = self.set_loop_stage
          # debug when configuring a plugin
          debug_config_changes = self.config_data.get('PLUGINS_DEBUG_CONFIG_CHANGES', False) # Ugly but needed
          # end debug
          
          _module_version = _config_dict.get('MODULE_VERSION', '0.0.0')
          
          plugin = _cls_def(
            log=self.log,
            global_shmem=self.shmem, # this SHOULD NOT be used for inter-plugin mem access
            plugins_shmem=self.plugins_shmem,
            stream_id=stream_name,
            signature=signature,
            default_config=_config_dict,
            upstream_config=upstream_config,
            environment_variables=self._environment_variables,
            initiator_id=initiator_id,
            initiator_addr=initiator_addr,
            session_id=session_id,
            threaded_execution_chain=self._run_on_threads,
            payloads_deque=self.comm_shared_memory['payloads'][instance_hash],
            commands_deque=self.comm_shared_memory['commands'][instance_hash],
            ee_ver=self.owner.__version__,
            runs_in_docker=self.owner.runs_in_docker,
            docker_branch=self.owner.docker_source,
            debug_config_changes=debug_config_changes,
            version=_module_version,
            pipelines_view_function=self.owner.get_pipelines_view,
            pipeline_use_local_comms_only=self._dct_config_streams[stream_name].get(ct.CONFIG_STREAM.K_USE_LOCAL_COMMS_ONLY, False),
          )
          if plugin.cfg_runs_only_on_supervisor_node:
            if not self.is_supervisor_node:
              self.P(
                "Plugin {}:{} runs ONLY on supervisor node. Skipping.".format(signature, instance_id), 
                color='r', boxed=True,
              )
              plugin = None
              # continue
            else:
              self.P("Plugin {}:{} runs only on supervisor node. Running.".format(signature, instance_id), color='g')
          # endif runs only on supervisor node
          self.set_loop_stage('2.bm.refresh.new_instance_done: {}:{}:{}'.format(stream_name, signature, instance_id))
        except Exception as exc:
          plugin = None
          trace = traceback.format_exc()
          msg = "Plugin init FAILED for business plugin {} instance {}".format(signature, instance_id)
          info = str(exc)
          if "validating" not in info:
            info += '\n' + trace
          self.P(msg + ': ' + info, color='r')
          self._create_notification(
            notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
            msg=msg,
            signature=signature,
            instance_id=instance_id,
            stream_name=stream_name,
            info=info,
            displayed=True,
          )
        #end try-except

        self._dct_current_instances[instance_hash] = plugin

        if plugin is None:
          continue

        self.P("New plugin instance {} added for exec.".format(plugin), color='g')
        if self._run_on_threads:
          plugin.start_thread()
      #endif new instance
      else:
        # I do have the instance, I just need to modify the config
        plugin = self._dct_current_instances[instance_hash]
        if plugin is not None:
          # next we need to check if the config has changed and handle also the particular
          # case when the plugin just received a INSTANCE_COMMAND
          plugin.maybe_update_instance_config(
            upstream_config=upstream_config,
            session_id=session_id,
            modified_by_addr=modified_by_addr,
            modified_by_id=modified_by_id,
          )
          self.set_loop_stage('2.bm.refresh.maybe_update_instance_config.DONE: {}:{}:{}'.format(stream_name, signature, instance_id))
        #endif
      #endif

    return current_instances

  def _stop_plugin(self, instance_hash):
    plugin = self._dct_current_instances[instance_hash]

    if plugin is None:
      return

    if self._run_on_threads:
      plugin.stop_thread()

    plugin = self._dct_current_instances.pop(instance_hash)
    self._dct_stop_timings.pop(instance_hash, None)
    del plugin
  
    sleep(0.1)
    gc.collect()
    return

  def maybe_stop_finished_plugins(self):
    """
    This method will check if any plugins that have been marked for deletion are
    are still working and forces the closing
    """
    max_stop_lag_time = self.log.config_data.get('MAX_PLUGIN_STOP_TIME', 60)
    instances = list(self._dct_stop_timings.keys())
    for instance_hash in instances:
      if instance_hash in self._dct_stop_timings:
        stop_init_time = self._dct_stop_timings[instance_hash]
        elapsed = time() - stop_init_time
        plugin = self._dct_current_instances.get(instance_hash)
        if elapsed > max_stop_lag_time and plugin is not None:
          self.P(ct.BM_PLUGIN_END_PREFIX + "Stopping lagged ({:.1f}s/{:.1f}s) pugin {}".format(
            elapsed, max_stop_lag_time, plugin
            ), color='r'
          )
          self._stop_plugin(instance_hash)
    return

  def _send_stop_signal_to_plugin(self, instance_hash, forced=False):
    plugin = self._dct_current_instances.get(instance_hash)
    if plugin is None:
      self.P("Received STOP for unvail plugin instance {}".format(instance_hash), color='r')
      if instance_hash in self._dct_current_instances:
        self.P("  Deleting {} from current instances".format(instance_hash), color='r')
        del self._dct_current_instances[instance_hash]
      if instance_hash in self._graceful_stop_instances:
        self.P("  Deleting {} from instances marked for graceful stop".format(instance_hash), color='r')
        del self._graceful_stop_instances[instance_hash]
      return

    if plugin.done_loop:
      self.P("Received STOP for already stopped {}:{}".format(instance_hash, plugin), color='r')
      # TODO: maybe add a sleep so that we do not spam this thing
      self._graceful_stop_instances.pop(instance_hash, 0)
      return

    if instance_hash not in self._dct_stop_timings:
      self._dct_stop_timings[instance_hash] = time()

    nr_inputs = len(plugin.upstream_inputs_deque)

    if (nr_inputs == 0) or forced:
      self.P(ct.BM_PLUGIN_END_PREFIX + "Stopping {}".format(plugin), color='y')
      self._stop_plugin(instance_hash)
      nr_graceful_tries = self._graceful_stop_instances.pop(instance_hash, 0)
      self.P("Stopped `{}` (pend.inp:{}, forced:{}, graceful cnt.:{})".format(repr(plugin), nr_inputs, forced, nr_graceful_tries), color='y')
    else:
      self._graceful_stop_instances[instance_hash] += 1
      if self._graceful_stop_instances[instance_hash] == 1:
        self.P(ct.BM_PLUGIN_END_PREFIX + "Gracefull marking for stopping {}".format(plugin), color='y')
        self.P("Marked `{}` to be gracefully stopped as it has {} pending inputs. Waiting to consume all pending inputs.".format(
          repr(plugin), nr_inputs),
          color='r'
        )
    #endif

    return

  def close(self):
    self.P("Stopping all business plugins...", color='y')
    keys = list(self._dct_current_instances.keys())
    for _hash in keys:
      self._send_stop_signal_to_plugin(instance_hash=_hash, forced=True)
    self.P("Done stopping all business plugins.", color='y')
    return

  def _deallocate_unused_instances(self, current_instances):
    keys = list(self._dct_current_instances.keys())
    for _hash in keys:
      if _hash not in current_instances:
        self._send_stop_signal_to_plugin(instance_hash=_hash, forced=False)

    keys = list(self._graceful_stop_instances.keys())
    for _hash in keys:
      self._send_stop_signal_to_plugin(instance_hash=_hash, forced=False)
    return

  @property
  def dct_instances_details(self):
    dct_instances_details = {}
    for instance_hash, plugin in self._dct_current_instances.items():
      if plugin is None:
        continue
      dct_instances_details[instance_hash] = (
        plugin.get_stream_id(),
        plugin.get_signature(),
        plugin.get_instance_config()
      )
    return dct_instances_details

  def fetch_ai_engines(self):
    self.dct_serving_processes_details = {}
    self.dct_serving_processes_startup_params = {}
    currently_used_ai_engines = set()

    for instance_hash, (stream_id, signature, instance_config) in self.dct_instances_details.items():
      plugin = self._dct_current_instances[instance_hash]
      ai_engine = plugin.cfg_ai_engine
      if ai_engine is None:
        continue

      # this config params go into serving process overall inputs
      inference_ai_engine_params = instance_config.get('INFERENCE_AI_ENGINE_PARAMS', {})

      # this is used only at model startup
      startup_ai_engine_params = instance_config.get('STARTUP_AI_ENGINE_PARAMS', {})

      ### for 'AI_ENGINE' there can be either single value or multiple values (list)
      ### For the second case we expect that for 'INFERENCE_AI_ENGINE_PARAMS' and 'STARTUP_AI_ENGINE_PARAMS'
      ### to (maybe) have params for each model serving process in the list.
      assert isinstance(ai_engine, (str, list))
      if not isinstance(ai_engine, list):
        ai_engine = ai_engine.lower()
        tmp_inference_ai_engine_params = {}
        tmp_startup_ai_engine_params = {}
        if ai_engine not in inference_ai_engine_params:
          tmp_inference_ai_engine_params[ai_engine] = inference_ai_engine_params
          inference_ai_engine_params = tmp_inference_ai_engine_params
        #endif

        if ai_engine not in startup_ai_engine_params:
          tmp_startup_ai_engine_params[ai_engine] = startup_ai_engine_params
          startup_ai_engine_params = tmp_startup_ai_engine_params
        #endif not in startup

        ai_engine = [ai_engine]
      #endif is just a string

      for _ai_engine in ai_engine:
        _ai_engine = _ai_engine.lower()
        inference_params = inference_ai_engine_params.get(_ai_engine, {})
        startup_params = startup_ai_engine_params.get(_ai_engine, {})
        t = (stream_id, json.dumps(inference_params))
        model_instance_id = startup_params.get('MODEL_INSTANCE_ID', None)
        if model_instance_id is not None:
          key = (_ai_engine, model_instance_id)
        else:
          key = _ai_engine

        if key not in self.dct_serving_processes_details:
          self.dct_serving_processes_details[key] = {}
        if t not in self.dct_serving_processes_details[key]:
          self.dct_serving_processes_details[key][t] = []
        if key not in currently_used_ai_engines:
          currently_used_ai_engines.add(key)

        self.dct_serving_processes_details[key][t].append(instance_hash)

        # TODO check current startup_params overwrite other startup_params
        # for _model_serving, which is a case of misconfiguration
        self.dct_serving_processes_startup_params[key] = startup_params
      #endfor
    #endfor

    return currently_used_ai_engines


  @property
  def any_overloaded_plugins(self):
    overloaded_plugins = self.get_overloaded_plugins()
    return len(overloaded_plugins) > 0

  def get_overloaded_plugins(self):
    all_instance_hashes = list(self._dct_current_instances)
    lst_overloaded = []
    for instance_hash in all_instance_hashes:
      plugin = self._dct_current_instances.get(instance_hash)
      if plugin is not None and plugin.is_queue_overflown:
        overflow = plugin.is_queue_overflown
        qsize = plugin.input_queue_size 
        qsizemax = plugin.cfg_max_inputs_queue_size
        status = "{}/{}".format(qsize, qsizemax)
        lst_overloaded.append(
          (plugin.get_stream_id(), plugin.get_signature(), plugin.get_instance_id(), overflow, status)
        )
    return lst_overloaded

  def get_plugin_default_config(self, signature):
    _module_name, _class_name, _cls_def, _config_dict = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_BIZ_PLUGINS,
      name=signature,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_BIZ_PLUGINS,
      verbose=0,
      safety_check=True, # perform safety check on custom biz plugins # TODO: should we do this?
      safe_locations=ct.PLUGIN_SEARCH.SAFE_BIZ_PLUGINS,
      safe_imports=ct.PLUGIN_SEARCH.SAFE_BIZ_IMPORTS
    )

    return _config_dict


  def execute_all_plugins(self, dct_business_inputs):
    self.log.start_timer('execute_all_business_plugins')
    all_instance_hashes = list(dct_business_inputs.keys())
    for instance_hash in all_instance_hashes:
      inputs = dct_business_inputs.get(instance_hash)
      plugin = self.get_subaltern(instance_hash) # this or `self._dct_current_instances[instance_hash]`
      if plugin is None:
        dct_info = self._dct_instance_hash_log.get(instance_hash, {})
        stream_name = dct_info.get(ct.PAYLOAD_DATA.STREAM_NAME)
        signature = dct_info.get(ct.PAYLOAD_DATA.SIGNATURE)
        instance_id = dct_info.get(ct.PAYLOAD_DATA.INSTANCE_ID)
        initiator_id = dct_info.get(ct.PAYLOAD_DATA.INITIATOR_ID)
        session_id = dct_info.get(ct.PAYLOAD_DATA.SESSION_ID)
        msg = "Biz plugin instance execution error"
        info = "Biz plugin instance '{}' was deleted yet the pipeline has not safely cleaned the dataflow. Deleted instance data: {}".format(
          instance_hash, dct_info,
        )
        self.P(info, color='r')
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
          msg=msg,
          info=info,
          stream_name=stream_name,
          signature=signature,
          instance_id=instance_id,
          initiator_id=initiator_id,
          session_id=session_id,
          displayed=True,
        )
        continue

      if not self._run_on_threads:
        # postponing stuff
        if plugin.is_process_postponed:
          # if process needs postponing just do not add new inputs to process
          # and if it runs in parallel the loop will check this
          continue

        if plugin.is_outside_working_hours:
          # if process is outside working hours again do not add inpus
          # and if it runs in parallel the loop will check this
          continue
        # end postponing stuff
        
        plugin.add_inputs(inputs)

        plugin.execute()
      else:
        # if the process is running (default) on thread we just need to add
        # data to its inputs queue
        plugin.add_inputs(inputs)
      #endif

    #endfor
    self.log.stop_timer('execute_all_business_plugins', skip_first_timing=False)
    return




#global dependencies
import itertools

#local dependencies
from naeural_core import Logger

CONFIG = {
    "CAP_RESOLUTION": 50,
    "INITIATOR_ID": "1111",
    "LIVE_FEED": False,
    "NAME": None,
    "PLUGINS": [
        {
            "INSTANCES": [
                {
                    "AI_ENGINE": [
                        "general_detector"
                    ],
                    "COORDS": "NONE",
                    "INSTANCE_ID": "OBJ_DET0",
                    "OBJECT_TYPE": [
                        "person"
                    ]
                }
            ],
            "SIGNATURE": "OBJECT_DETECTOR_TRACKER_01"
        }
    ],
    "RECONNECTABLE": False,
    "STREAM_CONFIG_METADATA": {
        "NOTIFY_DOWNLOAD_DOWNSTREAM": True,
        "C_ARRAY": True
    },
    "STREAM_WINDOW": 1,
    "TYPE": "VideoFile",
    "URL": "https://www.dropbox.com/s/mcp9dk82jdcr6t9/Camera%20bariere%20acces%20Kaufland%201.mp4?dl=1"
}

if __name__ == '__main__':
  log = Logger(
    lib_name='SBTST',
    base_folder='.',
    app_folder='_local_cache',
    config_file='config_startup.txt',
    max_lines=1000, 
    TF_KERAS=False
  )
  
  AI_ENGINE = ['general_detector', 'advanced_general_detector']
  STREAM_WINDOW = [1, 32]
  C_ARRAY = [True, False]
  
  for comb in itertools.product(AI_ENGINE, STREAM_WINDOW, C_ARRAY):
    ai_eng, s_win, c_arr = comb
    
    cfg = CONFIG.copy()
    name = '_'.join([str(x) for x in comb])
    cfg['NAME'] = name
    cfg['PLUGINS'][0]['INSTANCES'][0]['AI_ENGINE'] = ai_eng
    cfg['STREAM_CONFIG_METADATA']['C_ARRAY'] = c_arr
    cfg['STREAM_WINDOW'] = s_win
    
    log.save_json(cfg, 'xperimental/data_capture/{}.txt'.format(name))
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  


import os
from naeural_core import constants as ct


class _ValidateConfigStartupMixin(object):
  def __init__(self):
    super(_ValidateConfigStartupMixin, self).__init__()
    return

  def validate_config_retrieve(self):
    config_retrieve = self.config_data.get('CONFIG_RETRIEVE', None)
    app_config_idx = None
    streams_config_idx = []

    if config_retrieve is None:
      msg = "'{}' is not configured in `config_startup.txt`. The application will start using the default value. Pay attention that this parameter defines the source of the configurations (app and streams).".format('CONFIG_RETRIEVE')
      self.add_warning(msg)
    else:
      if not isinstance(config_retrieve, list):
        msg = "'{}' is not well configured in `config_startup.txt`. The value for it should be List[Dict].".format('CONFIG_RETRIEVE')
        self.add_error(msg)
        return
      #endif

      if len(config_retrieve) == 0:
        msg = "'{}' is not configured in `config_startup.txt`. The application will start using the default value. Pay attention that this parameter defines the source of the configurations (app and streams).".format('CONFIG_RETRIEVE')
        self.add_warning(msg)
      else:
        for idx,elem in enumerate(config_retrieve):
          if not isinstance(elem, dict):
            msg = "'{}' is not well configured in `config_startup.txt`. The value for it should be List[Dict]. Please check '{}'".format('CONFIG_RETRIEVE', elem)
            self.add_error(msg)
          else:
            if 'TYPE' not in elem:
              msg = "'{}' is not well configured in `config_startup.txt`. Each dictionary should have 'TYPE' configured. Please check '{}'".format('CONFIG_RETRIEVE', elem)
              self.add_error(msg)
            #endif

            if 'APP_CONFIG_ENDPOINT' in elem:
              app_config_idx = idx

            if 'STREAMS_CONFIGS_ENDPOINT' in elem:
              streams_config_idx.append(idx)
        #endfor
      #endif
    #endif

    fld_box_config = self.log.get_data_subfolder('box_configuration')
    content_box_config = []
    if fld_box_config is not None:
      content_box_config = os.listdir(fld_box_config)

    if app_config_idx is None and 'config_app.txt' not in content_box_config:
      msg = "No 'APP_CONFIG_ENDPOINT' configured in 'CONFIG_RETRIEVER' in `config_startup.txt` and there is no `config_app.txt` in `_data/box_configuration`."
      self.add_error(msg)

    return



import os
import numpy as np

class _TimeseriesBenchmakerMixin(object):
  """
  Mixin for timeseries benchmarker functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_MatplotlibMixin`:
    - self.add_copyright_to_plot
  """

  def __init__(self):
    super(_TimeseriesBenchmakerMixin, self).__init__()

    try:
      from .matplotlib_mixin import _MatplotlibMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _TimeseriesBenchmakerMixin without having _MatplotlibMixin")

    self.timeseries_benchmarker = None
    self.bp = None
    return

  def _setup_benchmarker(self, **kwargs):
    try:
      from naeural_core.local_libraries.ts import TimeseriesBenchmarker
    except:
      from ts import TimeseriesBenchmarker

    self.timeseries_benchmarker = TimeseriesBenchmarker(log=self, **kwargs)
    self.bp = self.timeseries_benchmarker
    return

  def supress_benchmarker_messages(self):
    self.P("Supressing BP messages!")
    self.timeseries_benchmarker.DEBUG = False
    return

  def get_baseline_desc(self, **kwargs):
    return self.timeseries_benchmarker.get_baseline_desc(**kwargs)

  def compare_timeseries_results(self, **kwargs):
    return self.timeseries_benchmarker.compare_timeseries_results(**kwargs)

  def multiperiod_compare_timeseries_results(self, **kwargs):
    return self.timeseries_benchmarker.multiperiod_compare_timeseries_results(**kwargs)

  def calc_timeseries_error(self, **kwargs):
    return self.timeseries_benchmarker.calc_timeseries_error(**kwargs)

  def autoregression(self, **kwargs):
    return self.timeseries_benchmarker.autoregression(**kwargs)

  def start_autoregression_benchmark(self, **kwargs):
    return self.timeseries_benchmarker.start_autoregression_benchmark(**kwargs)

  def add_autoregression_benchmark(self, **kwargs):
    return self.timeseries_benchmarker.add_autoregression_benchmark(**kwargs)

  def get_autoregression_benchmark(self, **kwargs):
    return self.timeseries_benchmarker.get_autoregression_benchmark(**kwargs)

  def get_autoregression_model_history(self, **kwargs):
    return self.timeseries_benchmarker.get_autoregression_model_history(**kwargs)

  def get_baseline_prediction(self, **kwargs):
    return self.timeseries_benchmarker.get_baseline_prediction(**kwargs)

  def debug_series(self, **kwargs):
    return self.timeseries_benchmarker.debug_series(**kwargs)

  def autoregression_baselines(self, **kwargs):
    return self.timeseries_benchmarker.autoregression_baselines(**kwargs)

  def get_model_result_dict(self, **kwargs):
    return self.timeseries_benchmarker._get_model_result_dict(**kwargs)

  def show_series(self,
                  dct_series,
                  reality=None,
                  title='Series',
                  xticks=None,
                  save_png=True,
                  vertical_lines=None,
                  style='-'
                  ):
    """
    Parameters
    ----------
    dct_series : dict
      dictionary with each series (prediction if reality is also added).

    reality : ndarray or dict, optional
      vector with reality values to be plotted against each series. The default is None.

    title : str, optional
      name of the series. The default is 'Series'.

    xticks : TYPE, optional
      labels of the series. The default is None.

    vertical_lines: list, optional
      list of x positions where to draw vertical lines

    Returns
    -------
    Name of the png file.

    """
    import matplotlib.pyplot as plt
    plt.style.use('ggplot')
    names = sorted(list(dct_series.keys()))
    values = {}
    nr_steps = dct_series[names[0]].shape[0]
    nr_series = len(dct_series)
    self.P("Plotting {} series with {} steps {}".format(
      nr_series,
      nr_steps,
      "with reality" if reality is not None else "without reality/baseline"))
    if reality is not None:
      assert type(reality) in [dict, list, np.ndarray], "reality must be a vector or dict"
    if type(reality) in [list, np.ndarray]:
      if type(reality) == np.ndarray:
        assert len(reality.shape) == 1, "reality must be a vector"
      dct_reality = {n: reality for n in names}
    else:
      dct_reality = reality

    for name in names:
      vals = dct_series[name]
      assert type(vals) in [np.ndarray, list], "the dictionary must contain vectors"
      vals = np.array(vals)
      assert len(vals.shape) == 1, "the dictionary must contain vectors"
      assert vals.shape[0] == nr_steps, "all series must have same size"
      values[name] = vals

    vlines = []
    if xticks is not None:
      assert len(xticks) == nr_steps, "x labels number different from series nr of steps"
      assert type(xticks[0]) == str, "x labels must be list of strings"
      vlines = np.argwhere(np.array(xticks) == 'Su').ravel()
      _xticks = xticks
    else:
      _xticks = [str(x) for x in range(1, nr_steps + 1)]

    _w = 13
    _h = 8 * nr_series
    fig, ax = plt.subplots(nr_series, 1, figsize=(_w, _h))

    if dct_reality is not None:
      for real_name in dct_reality:
        if real_name not in values:
          raise ValueError("Something is wrong. Reality series '{}' not found in series {}".format(
            real_name, list(values.keys())))

    for i, name in enumerate(values):
      vals = values[name]
      ax[i].plot(range(nr_steps), vals, 'b' + style, label=name)
      max_steps = nr_steps
      if dct_reality is not None:
        _real = dct_reality.get(name)
        if _real is not None:
          max_steps = max(max_steps, len(_real))
          ax[i].plot(np.arange(len(_real)), _real, 'g' + style, label='reality')
      if xticks is not None:
        ax[i].set_xticks(range(max_steps))
        ax[i].set_xticklabels(_xticks)
      ax[i].set_title('Series: ' + name)
      for vl in vlines:
        ax[i].axvline(vl, color='r')
      if vertical_lines is not None:
        if type(vertical_lines) == int:
          vertical_lines = [vertical_lines]
        for vl2 in vertical_lines:
          ax[i].axvline(vl2, color='k', linestyle='--')
      ax[i].legend()

    fig.suptitle(title, fontsize=20)

    self.add_copyright_to_plot(ax[i])

    fig.tight_layout()
    offset = max(0, (3 - nr_series) * 0.05)
    fig.subplots_adjust(top=0.95 - offset)
    if save_png:
      fn = os.path.join(self._outp_dir, title + '.png')
      plt.savefig(fn)
      self.P("Saved '{}'".format(fn))
    plt.show()
    return


import os
import shutil
import subprocess
import tempfile
import psutil

from jinja2 import Environment, FileSystemLoader

from naeural_core.business.base import BasePluginExecutor
from naeural_core.business.mixins_libs.ngrok_mixin import _NgrokMixinPlugin

__VER__ = '0.0.0.0'

# This may be used for running ready-made builds
WEB_APP_DEFAULT_ENV_VARS = {
  "HOSTNAME": "0.0.0.0"
}
WEB_APP_DEFAULT_GIT_TRACK_MODE = 'commit'

_CONFIG = {
  **BasePluginExecutor.CONFIG,
  'ALLOW_EMPTY_INPUTS': True,
  'RUN_WITHOUT_IMAGE': True,
  'PROCESS_DELAY': 5,
  
  'GIT_REQUEST_DELAY': 60 * 10,  # 10 minutes

  'NGROK_USE_API': True,
  'NGROK_ENABLED': False,
  'NGROK_DOMAIN': None,
  'NGROK_EDGE_LABEL': None,
  'NGROK_URL_PING_INTERVAL': 30,

  'NGROK_AUTH_TOKEN': None,
  
  'SUPRESS_LOGS_AFTER_INTERVAL' : 120,  # 2 minutes until logs suppression

  'ASSETS': None,

  'SETUP_COMMANDS': [],
  'START_COMMANDS': [],
  'ENV_VARS': {},
  'AUTO_START': True,
  'FORCED_RELOAD_INTERVAL': None,

  'PORT': None,
  "DEBUG_WEB_APP": False,

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES']
  },
}


class BaseWebAppPlugin(_NgrokMixinPlugin, BasePluginExecutor):
  """
  A base plugin which will handle the lifecycle of a web application.
  Through this plugin, you can expose your business logic as a web application,
  using some implementation of a web server.

  You can also deploy your web application to the internet using ngrok.
  To do this, set the `NGROK_ENABLED` flag to True in config and set the necessary
  environment variables.

  TODO: add ngrok necessary data in the config (after securing the configs)
  """

  CONFIG = _CONFIG
  
  def _on_init(self):

    self.__git_commit_hash = None
    self.__last_release_tag = None
    self.__last_asset_id = None
    self.__last_asset_updated_at = None
    self.__git_request_time = 0
    self.__first_log_displayed = None
    
    self._reset_ngrok()
    
    self.__last_ngrok_url_ping_ts = 0
    self.__wait_count = 0

    # TODO: move this to process
    self.__allocate_port()

    self.prepared_env = None
    self.base_env = None

    self.__init_temp_dir()
        
    self.assets_initialized = False
    self.failed = False
    self.__commands_ready = False

    self.can_run_start_commands = self.cfg_auto_start
    self.setup_commands_started = []
    self.setup_commands_finished = []
    self.setup_commands_processes = []
    self.setup_commands_start_time = []

    self.start_commands_started = []
    self.start_commands_finished = []
    self.start_commands_processes = []
    self.start_commands_start_time = []
    self.webapp_reload_last_timestamp = 0

    self.dct_logs_reader = {}
    self.dct_err_logs_reader = {}

    super(BaseWebAppPlugin, self)._on_init()
    return
  
  
  def __init_temp_dir(self):
    if getattr(self, "script_temp_dir", None) is not None:
      self.P(f"Deleting {self.script_temp_dir} ...")
      shutil.rmtree(self.script_temp_dir)
    self.script_temp_dir = tempfile.mkdtemp()
    self.P(f"Created {self.script_temp_dir}")
    return
  
  
  def __can_request_git(self):
    if self.time() - self.__git_request_time > self.cfg_git_request_delay:
      self.__git_request_time = self.time()
      return True
    return False


  # Port allocation
  def __check_port_valid(self):
    # Check the config as we're going to use it to start processes.
    port = self.get_port()
    if not isinstance(port, int):
      raise ValueError("Port not an int")
    if port < 0 or port > 65535:
      raise ValueError("Invalid port value {}".format(port))
    return

  def __get_all_used_ports(self):
    res = set()
    for conn in psutil.net_connections(kind='all'):
      # Local address is not always a tuple.
      if not isinstance(conn.laddr, str):
        res.add(conn.laddr.port)  # Local port
    # endfor
    return sorted(res)

  def __allocate_port(self):
    """
    TODO: (Bleo) this should be migrated to the new correct version 
    
    
    In case the port is not provided, a random unused one will be allocated.
    In case the port is provided, it will be checked for availability.
    If the provided port is not available, the allocation will be retried
    until a port is found.
    """
    cnt_tries = 0
    done = False
    while not done:
      with self.managed_lock_resource('USED_PORTS'):
        if 'USED_PORTS' not in self.plugins_shmem:
          self.plugins_shmem['USED_PORTS'] = {}
        dct_shmem_ports = self.plugins_shmem['USED_PORTS']
        used_ports = self.__get_all_used_ports()
        configured_port = self.get_port()
        if configured_port is not None:
          self.__check_port_valid()

          if configured_port not in used_ports:
            dct_shmem_ports[self.str_unique_identification] = configured_port
            done = True
          else:
            cnt_tries += 1
        else:
          port = self.np.random.randint(30000, 32500)
          total_tries = 1000
          tries = 0
          while port in used_ports and tries < total_tries:
            tries += 1
            port = self.np.random.randint(30000, 32500)
          # endwhile
          if tries >= total_tries:
            raise Exception(f"Could not find an available port after {total_tries} tries.")
          # endif tries
          dct_shmem_ports[self.str_unique_identification] = port
          done = True
        # endif port
      # endwith lock
      if not done:
        sleep_seconds = 5
        self.P(
          f"Preconfigured port {configured_port} is already in use at try {cnt_tries}. Retrying in {sleep_seconds}...",
          color='r'
        )
        self.sleep(sleep_seconds)
    return

  def __deallocate_port(self):
    # TODO: both this and __allocate_port don t actually allocate/deallocate ports
    #  They just mark them as so.
    port = self.port

    with self.managed_lock_resource('USED_PORTS'):
      if 'USED_PORTS' in self.plugins_shmem:
        self.plugins_shmem['USED_PORTS'].pop(self.str_unique_identification, None)
    # endwith lock

    self.P(f"Released port {port}")
    return

  def get_additional_env_vars(self):
    return {
      "EE_NODE_ALIAS": self.eeid,
      "EE_NODE_ADDRESS": self.bc.address,
      "EE_NODE_ETH_ADDRESS": self.bc.eth_address,
      "EE_NODE_NETWORK": self.evm_network,
      "EE_NODE_VER": self.ee_ver
    }

  def __prepare_env(self, assets_path):
    # pop all `EE_` keys
    prepared_env = dict(self.os_environ)
    to_pop_keys = []
    for key in prepared_env:
      if key.startswith('EE_'):
        to_pop_keys.append(key)
    # endfor all keys

    for key in to_pop_keys:
      prepared_env.pop(key)

    # add mandatory keys
    prepared_env["PWD"] = self.script_temp_dir
    prepared_env["PORT"] = str(self.port)

    self.base_env = prepared_env.copy()

    # add default keys
    prepared_env = {
      **prepared_env,
      **WEB_APP_DEFAULT_ENV_VARS,
    }

    # add optional keys, found in `.env` file from assets folder if provided.
    if assets_path is not None:
      env_file_path = self.os_path.join(assets_path, '.env')
      if self.os_path.exists(env_file_path):
        with open(env_file_path, 'r') as f:
          for line in f:
            if line.startswith('#') or len(line.strip()) == 0:
              continue
            key, value = line.strip().split('=', 1)
            prepared_env[key] = value
        # endwith

        # TODO: should we remove the .env file?
      # endif env file
    # endif assets path provided

    # add optional keys, found in config
    if self.cfg_env_vars is not None and isinstance(self.cfg_env_vars, dict):
      processed_env = {k: str(v) for k, v in self.cfg_env_vars.items()}
      prepared_env.update(processed_env)

    additional_env_vars = self.get_additional_env_vars() or {}
    prepared_env.update(additional_env_vars)

    self.prepared_env = prepared_env

    return prepared_env

  # process handling methods
  def __run_command(self, command, env=None, read_stdout=True, read_stderr=True):
    if isinstance(command, list):
      command_args = command
    elif isinstance(command, str):
      command_args = command.split(' ')
    else:
      raise ValueError("Command must be a string or a list of strings")

    process = subprocess.Popen(
      command_args,
      env=env,
      cwd=self.script_temp_dir,
      stdout=subprocess.PIPE if read_stdout else None,
      stderr=subprocess.PIPE if read_stderr else None,
      bufsize=0, # this is important for real-time output
    )
    logs_reader = self.LogReader(process.stdout, size=100) if read_stdout else None
    err_logs_reader = self.LogReader(process.stderr, size=100) if read_stderr else None
    return process, logs_reader, err_logs_reader

  def __wait_for_command(self, process, timeout):
    process_finished = False
    failed = False
    try:
      process.wait(timeout)
      failed = process.returncode != 0
      process_finished = True
    except subprocess.TimeoutExpired:
      pass

    return process_finished, failed

  def kill_process_and_children(self, proc: subprocess.Popen, timeout=3):
    try:
      parent = psutil.Process(proc.pid)
    except psutil.NoSuchProcess:
      return

    # Get child processes (recursive=True → grandchildren, etc.)
    children = parent.children(recursive=True)

    # Send SIGKILL to children first
    for child in children:
      try:
        child.kill()
      except psutil.NoSuchProcess:
        pass

    # Kill the parent
    try:
      parent.kill()
    except psutil.NoSuchProcess:
      pass

    # Wait for them to terminate
    gone, alive = psutil.wait_procs([parent, *children], timeout=timeout)
    if alive:
      # Some processes are still alive
      for p in alive:
        p.terminate()  # or p.kill() again
    return

  def __maybe_kill_process(self, process, key, max_tries=5):
    if process is None:
      return
    tries = 0
    success = False
    while tries < max_tries and not success:
      tries += 1
      try:
        self.P(f"Forcefully killing process {key}(try {tries}/{max_tries})")
        self.kill_process_and_children(process, timeout=3)
        # Check if it's done
        if process.poll() is not None:
          self.P(f"Process {key} is fully terminated.")
          success = True
        else:
          self.P(f"Process {key} still alive. Retrying.")
        self.__maybe_print_key_logs(key)
        if success:
          self.P(f"Killed process {key} from {tries} tries.")
      except Exception as exc:
        self.P(f'Could not kill process {key} (try {tries}/{max_tries}). Reason: {exc}')
    # endwhile
    return


  def on_log_handler(self, text):
    return


  # logs handling methods
  def __maybe_print_all_logs(self, indent=35):
    for key, logs_reader in self.dct_logs_reader.items():
      if logs_reader is not None:
        logs = logs_reader.get_next_characters()
        if len(logs) > 0:
          self.on_log_handler(logs)
          if isinstance(self.cfg_supress_logs_after_interval, int) and self.cfg_supress_logs_after_interval > 0:
            time_since_first_log = 0 if self.__first_log_displayed is None else (self.time() - self.__first_log_displayed)
            if time_since_first_log > self.cfg_supress_logs_after_interval and len(logs) > 0:
              # only print logs if the first log was displayed less than `supress_logs_after_interval` seconds ago
              indented_logs = self.indent_strings(logs, indent=indent)
              self.P(f"Showing stdout logs [{key}]:\n{indented_logs}")
              self.logs.append(f"[{key}]: {logs}")
              if self.__first_log_displayed is None:
                self.__first_log_displayed = self.time()
            #end if time_since_first_log
          #end if supress_logs_after_interval
        # endif logs
      #end if logs_reader
    #endfor all logs readers

    for key, err_logs_reader in self.dct_err_logs_reader.items():
      if err_logs_reader is not None:
        err_logs = err_logs_reader.get_next_characters()
        if len(err_logs) > 0:
          self.on_log_handler(err_logs)
          indented_err_logs = self.indent_strings(err_logs, indent=indent)
          self.P(f"Showing error logs [{key}]:\n{indented_err_logs}")
          self.err_logs.append(f"[{key}]: {err_logs}")
    return


  def __maybe_print_key_logs(self, key):
    logs_reader = self.dct_logs_reader.get(key)
    if logs_reader is not None:
      logs = logs_reader.get_next_characters()
      if len(logs) > 0:
        self.P(f"[stdout][{key}]: {logs}")
        self.logs.append(f"[{key}]: {logs}")

    err_logs_reader = self.dct_err_logs_reader.get(key)
    if err_logs_reader is not None:
      err_logs = err_logs_reader.get_next_characters()
      if len(err_logs) > 0:
        self.P(f"[stderr][{key}]: {err_logs}")
        self.err_logs.append(f"[{key}]: {err_logs}")
    return


  def __get_delta_logs(self):
    logs = list(self.logs)
    logs = "".join(logs)
    self.logs.clear()

    err_logs = list(self.err_logs)
    err_logs = "".join(err_logs)
    self.err_logs.clear()

    return logs, err_logs

  def __maybe_read_and_stop_key_log_readers(self, key):
    if self.cfg_debug_web_app:
      self.P(f"Reading and stopping log readers for key {key}...")
    # endif debug web app
    logs_reader = self.dct_logs_reader.get(key)
    if logs_reader is not None:
      if self.cfg_debug_web_app:
        self.P(f"Stopping log reader for key {key}...")
      logs_reader.stop()

    err_logs_reader = self.dct_err_logs_reader.get(key)
    if err_logs_reader is not None:
      if self.cfg_debug_web_app:
        self.P(f"Stopping error log reader for key {key}...")
      err_logs_reader.stop()

    self.__maybe_print_key_logs(key)

    self.dct_logs_reader.pop(key, None)
    self.dct_err_logs_reader.pop(key, None)
    return

  def __maybe_read_and_stop_all_log_readers(self):
    self.P("Reading and stopping all log readers...")
    log_keys = set(list(self.dct_logs_reader.keys()) + list(self.dct_err_logs_reader.keys()))
    self.P(f"Log keys: {log_keys}")
    for key in log_keys:
      self.__maybe_read_and_stop_key_log_readers(key)
    return

  # setup commands methods
  def __maybe_run_all_setup_commands(self):
    if self.failed:
      return

    has_finished_setup_commands = self.__has_finished_setup_commands()
    if has_finished_setup_commands:
      return
    
    self.P("Running setup commands (`has_finished_setup_commands: {}`)".format(has_finished_setup_commands))

    for idx in range(len(self.get_setup_commands())):
      self.__maybe_run_nth_setup_command(idx)
    return


  def __maybe_close_setup_commands(self):
    if not any(self.setup_commands_started):
      self.P("No setup commands were started. Skipping teardown.")
      return

    if all(self.setup_commands_finished):
      self.P("All setup commands have finished. Skipping teardown.")
      return

    for idx, process in enumerate(self.setup_commands_processes):
      if self.setup_commands_started[idx] and not self.setup_commands_finished[idx]:
        self.P(f"Setup command nr {idx} has not finished. Killing it.")
        self.__maybe_kill_process(process, f"setup_{idx}")
    # endfor all setup commands
    return


  def __maybe_run_nth_setup_command(self, idx, timeout=None):
    if self.failed:
      return

    if idx > 0 and not self.setup_commands_finished[idx - 1]:
      # Previous setup command has not finished yet. Skip this one.
      return

    if not self.setup_commands_started[idx]:
      self.P(f"Running setup command nr {idx}: {self.get_setup_commands()[idx]}")
      proc, logs_reader, err_logs_reader = self.__run_command(self.get_setup_commands()[idx], self.base_env)
      self.setup_commands_processes[idx] = proc
      self.dct_logs_reader[f"setup_{idx}"] = logs_reader
      self.dct_err_logs_reader[f"setup_{idx}"] = err_logs_reader

      self.setup_commands_started[idx] = True
      self.setup_commands_start_time[idx] = self.time()
    # endif setup command started

    if not self.setup_commands_finished[idx]:
      finished, failed = self.__wait_for_command(
        process=self.setup_commands_processes[idx],
        timeout=0.1,
      )

      self.setup_commands_finished[idx] = finished

      if finished and not failed:
        self.__maybe_read_and_stop_key_log_readers(f"setup_{idx}")
        self.add_payload_by_fields(
          command_type="setup",
          command_idx=idx,
          command_str=self.get_setup_commands()[idx],
          command_status="success"
        )
        self.P(f"Setup command nr {idx} finished successfully")
      elif finished and failed:
        self.__maybe_read_and_stop_key_log_readers(f"setup_{idx}")
        self.add_payload_by_fields(
          command_type="setup",
          command_idx=idx,
          command_str=self.get_setup_commands()[idx],
          command_status="failed"
        )
        self.P(f"ERROR: Setup command nr {idx} finished with exit code {self.setup_commands_processes[idx].returncode}")
        self.failed = True
      elif not finished and timeout is not None and timeout > 0:
        if self.time() - self.setup_commands_start_time[idx] > timeout:
          self.setup_commands_processes[idx].kill()
          self.__maybe_read_and_stop_key_log_readers(f"setup_{idx}")
          self.P(f"ERROR: Setup command nr {idx} timed out")
          self.add_payload_by_fields(
            command_type="setup",
            command_idx=idx,
            command_str=self.get_setup_commands()[idx],
            command_status="timeout"
          )
          self.failed = True
    # endif setup command finished
    return


  def __has_finished_setup_commands(self):
    return all(self.setup_commands_finished)


  # start commands methods
  def __maybe_run_all_start_commands(self):
    if self.failed:
      return

    if self.__has_finished_start_commands():
      return

    if not self.__has_finished_setup_commands():
      return

    if not self.can_run_start_commands:
      return
    
    if not self.__all_start_running():
      self.P("Running START commands...")
    else:
      self.__wait_count += 1
      self.P(f"Waiting for START commands (pass #{self.__wait_count})...")
      self.sleep(1)

    for idx in range(len(self.get_start_commands())):
      self.__maybe_run_nth_start_command(idx)
    return


  def __all_start_running(self):
    result = True
    cmds = self.get_start_commands()
    for idx in range(len(cmds)):
      if not self.start_commands_started[idx]:
        result = False
        break
    return result


  def __maybe_close_start_commands(self):
    if not any(self.start_commands_started):
      self.P("Server was never started. Skipping teardown.")
      return

    for idx, process in enumerate(self.start_commands_processes):
      self.__maybe_kill_process(process, f"start_{idx}")
    # endfor all start commands

  def __maybe_run_nth_start_command(self, idx, timeout=5):
    if self.failed:
      return

    if idx > 0 and not self.start_commands_finished[idx - 1]:
      # Previous start command has not finished yet. Skip this one.
      return

    if not self.start_commands_started[idx]:
      cmd = self.get_start_commands()[idx]
      self.P(f"Running start command nr {idx}: {cmd}")
      proc, logs_reader, err_logs_reader = self.__run_command(cmd, self.prepared_env)
      self.start_commands_processes[idx] = proc
      self.dct_logs_reader[f"start_{idx}"] = logs_reader
      self.dct_err_logs_reader[f"start_{idx}"] = err_logs_reader

      self.start_commands_started[idx] = True
      self.start_commands_start_time[idx] = self.time()
    # endif start command started

    if not self.start_commands_finished[idx]:
      finished, _ = self.__wait_for_command(
        process=self.start_commands_processes[idx],
        timeout=0.1,
      )

      self.start_commands_finished[idx] = finished

      if finished:
        self.__maybe_read_and_stop_key_log_readers(f"start_{idx}")
        self.add_payload_by_fields(
          command_type="start",
          command_idx=idx,
          command_str=self.get_start_commands()[idx],
          command_status="failed"
        )
        self.P(f"Start command nr {idx} finished unexpectedly. Please check the logs.")
        self.failed = True

      elif self.time() - self.start_commands_start_time[idx] > timeout:
        self.start_commands_finished[idx] = True
        cmd = self.get_start_commands()[idx]
        self.add_payload_by_fields(
          command_type="start",
          command_idx=idx,
          command_str=cmd,
          command_status="success"
        )
        self.P(f"Start cmd {idx} '{cmd}' is running after {timeout}s.")
    # endif setup command finished
    return

  def __has_finished_start_commands(self):
    return all(self.start_commands_finished)
  
  
  def __reload_server(self):
    self.P("Initiating server reload...")
    
    self.__maybe_close_setup_commands()
    self.__maybe_close_start_commands()
    self.__maybe_read_and_stop_all_log_readers()

    self.assets_initialized = False
    self.failed = False

    self.setup_commands_started = [False] * len(self.get_setup_commands())
    self.setup_commands_finished = [False] * len(self.get_setup_commands())
    self.setup_commands_processes = [None] * len(self.get_setup_commands())
    self.setup_commands_start_time = [None] * len(self.get_setup_commands())

    self.start_commands_started = [False] * len(self.get_start_commands())
    self.start_commands_finished = [False] * len(self.get_start_commands())
    self.start_commands_processes = [None] * len(self.get_start_commands())
    self.start_commands_start_time = [None] * len(self.get_start_commands())

    self.__init_temp_dir()
    self.__deallocate_port()
    self.P('Attempting to init assets due to reload...')
    self.__allocate_port()
    # In case a new port is allocated instead of the one previously used.
    # The commands will be relogged in case something changed because of the
    # port reallocation.
    self.__commands_ready = False
    self.__maybe_setup_commands()
    self.__maybe_init_assets()
    return  
  
  
  
  def __check_new_repo_version(self):
    result = False  # return false by default including if no git url is provided
    if not isinstance(self.cfg_assets, dict):
      return result
    # endif valid assets dict
    operation = self.cfg_assets.get('operation')
    # first check not to run this operation too many times
    if operation == 'clone':
      url = self.cfg_assets.get('url')
      if url is None:
        self.P(f"Operation {operation}, but no URL provided. Check assets!", color="r")
        return result
      if not isinstance(url, str):
        self.P(f"Operation {operation}, but URL is not a string. Check assets!", color="r")
        return result
      # endif valid url provided
      username = self.cfg_assets.get('username')
      token = self.cfg_assets.get('token')
      track_mode = self.cfg_assets.get('track', WEB_APP_DEFAULT_GIT_TRACK_MODE)

      # If we haven't cloned anything yet, treat it like a new version:
      if self.__git_commit_hash is None and track_mode == 'commit':
        self.P("No previous local git commit info. Assuming new commit and initializing...")
        return True
      if self.__last_release_tag is None and track_mode == 'tag':
        self.P("No previous local release tag info. Assuming new release and initializing...")
        return True

      # 1. Check for a new commit.
      if track_mode == 'commit':
        if self.__can_request_git():
          remote_commit = self.git_get_last_commit_hash(repo_url=url, user=username, token=token)
          if remote_commit is None:
            self.P("Could not get the commit hash. Assuming no new version.")
            result = False
          elif remote_commit != self.__git_commit_hash:
            self.P(
              f"New git commit found: local={self.__git_commit_hash} vs remote={remote_commit}."
              f"Server reloading procedure will be initiated..."
            )
            result = True
        # endif can request git
      # 2. Check for a new release.
      elif track_mode == 'tag':
        if self.__can_request_git():
          last_release = self.git_get_last_release_tag(repo_url=url, user=username, token=token)
          if last_release is None:
            self.P("Could not get the last release tag. Assuming no new version.")
            result = False
          elif last_release != self.__last_release_tag:
            self.P(
              f"New release tag found: local={self.__last_release_tag} vs remote={last_release}."
              f"Server reloading procedure will be initiated..."
            )
            result = True
      # endif track_mode
    elif operation == 'release_asset':
      url = self.cfg_assets.get('url')
      if not isinstance(url, str):
        self.P(f"Operation release_asset, but URL is not valid. Check assets!", color="r")
        return result

      token = self.cfg_assets.get('token')
      release_tag_substring = self.cfg_assets.get('release_tag')
      asset_filter = self.cfg_assets.get('asset_filter')

      # If we have never downloaded anything, assume "new version"
      if self.__last_asset_id is None:
        self.P("[release_asset] No cached asset ID. Assuming new asset.")
        return True

      # Try to check for an updated asset
      if self.__can_request_git():  # or rename to __can_make_gh_api_request()
        new_info = self.git_get_latest_release_asset_info(
          repo_url=url,
          token=token,
          release_tag_substring=release_tag_substring,
          asset_filter=asset_filter
        )
        if not new_info:
          self.P("[release_asset] Could not get asset info. Assuming no new version.")
          return False

        # Compare new_info with our cached values
        new_asset_id = new_info["asset_id"]
        new_asset_updated = new_info["asset_updated_at"]
        old_asset_id = self.__last_asset_id
        old_asset_updated = self.__last_asset_updated_at

        if new_asset_id != old_asset_id or new_asset_updated != old_asset_updated:
          self.P(
            f"[release_asset] New asset found!\n"
            f"Old asset_id={old_asset_id}, updated_at={old_asset_updated}\n"
            f"New asset_id={new_asset_id}, updated_at={new_asset_updated}\n"
            "Server reloading procedure will be initiated..."
          )
          result = True
        # endif new asset version
      # endif can request git
    # endif operation handling
    return result

  def check_assets_path(self, assets_path, raise_exception=True):
    """
    Utility method for checking if the assets path is safe.
    That means it is either a URL or a safe relative path.
    Parameters
    ----------
    assets_path : str
        The path to the assets.
    raise_exception : bool
        If True, an exception will be raised if the path is not safe.
        Otherwise, an error will be added to the plugin.

    """
    # TODO: maybe allow some local assets in case of private networks?
    # If the assets path is a local path, it must be a safe path.
    if os.path.exists(assets_path) and not self.is_path_safe(assets_path, base_path='.'):
      msg = ("Local assets in not safe paths are off-limits!"
             "Please provide a URL or a safe relative path(without any symlink or \"../\").")
      if raise_exception:
        raise ValueError(msg)
      else:
        self.add_error(msg)
      return
    return

  def validate_assets(self):
    """
    Validate the configuration of `ASSETS`.
    This will check if someone is trying to retrieve local files or other sensitive data.

    Returns
    -------

    """
    # If no assets are provided, no validation is needed.
    if self.cfg_assets is None:
      return

    if not isinstance(self.cfg_assets, (dict, str)):
      self.add_error("`ASSETS`, if specified, must be a dictionary or a string.")
      return
    # endif valid assets type

    # If no assets path can be obtained, no validation is needed.
    assets_path = self.cfg_assets if isinstance(self.cfg_assets, str) else self.cfg_assets.get('url')
    if assets_path is None:
      return

    # The assets path must be a string.
    if not isinstance(assets_path, str):
      self.add_error("`ASSETS.url` must be a string.")
      return

    # If the assets path is a local path, it must be a safe path.
    self.check_assets_path(assets_path=assets_path, raise_exception=False)

    return

  # assets handling methods
  def __maybe_download_assets(self):
    """
      self.cfg_assets = {
        "url": "https://example.com/assets.zip",
        "operation": "download",
      }
      self.cfg_assets = {
        "url": "https://github.com/user/repo",
        "username": "username",
        "token": "token",
        "operation": "clone",
      }
      self.cfg_assets = {
        "url": "https://github.com/user/repo",
        "username": null,
        "token": null,
        "operation": "clone",
        "track": "commit"  # Default value
      }
      self.cfg_assets = {
        "url": "https://github.com/user/repo",
        "username": null,
        "token": null,
        "operation": "clone",
        "track": "tag"
      }
      self.cfg_assets = {
        "url": "https://github.com/user/repo",
        "username": null,
        "token": null,
        "operation": "clone",
        "track": "tag",
        "release_tag": "latest"
      }
      self.cfg_assets = {
        "url": "https://github.com/user/repo",
        "username": null,
        "token": null,
        "operation": "release_asset",
        "asset_filter": "your-web-app-build.zip"
        "release_tag": "latest"
      }
      self.cfg_assets = {
        "url": "/path/to/local/dir",
        "operation": "download",
      }
      self.cfg_assets = {
        "url"=[["base64_encoded_file", "encoded_file_name"], ...],
        "operation": "decode",
      }
    """
    # handle assets url: download, extract, then copy, then delete
    relative_assets_path = self.os_path.join('downloaded_assets', self.plugin_id, 'assets')

    # 1) Parse the assets configuration
    operation = None
    assets_path = None

    # check if assets is a dict or a string
    # and extract the url and operation
    if isinstance(self.cfg_assets, dict):
      dct_data = self.cfg_assets
      operation = dct_data.get("operation", "download")
      assets_path = dct_data.get("url", None)

    elif isinstance(self.cfg_assets, str):
      assets_path = self.cfg_assets
      operation = "download"

    if assets_path is None:
      return None
      # raise ValueError("No assets provided")

    self.check_assets_path(assets_path=assets_path, raise_exception=True)

    # 2) Handle each operation
    if operation == "release_asset":
      username = dct_data.get("username")
      token = dct_data.get("token")
      release_tag_substring = dct_data.get("release_tag")
      asset_filter = dct_data.get("asset_filter")

      download_dir = self.os_path.join(self.get_output_folder(), relative_assets_path)
      local_file = self.git_download_release_asset(
        repo_url=assets_path,
        user=username,
        token=token,
        release_tag_substring=release_tag_substring,
        asset_filter=asset_filter,
        download_dir=download_dir
      )
      if not local_file:
        err_msg = "[ERROR] Could not download the prebuilt artifact."
        self.P(err_msg, color='r')
        raise Exception(err_msg)
        return None

      # ------------------------------------
      # (A) Cache release metadata for checks
      # ------------------------------------
      # We can re-call git_get_latest_release_asset_info to get exactly what we used
      latest_info = self.git_get_latest_release_asset_info(
        assets_path, token=token,
        release_tag_substring=release_tag_substring,
        asset_filter=asset_filter
      )
      if latest_info:
        self.__last_release_tag = latest_info["release_tag"]
        self.__last_asset_id = latest_info["asset_id"]
        self.__last_asset_updated_at = latest_info["asset_updated_at"]
        self.P(f"Cached release tag: {self.__last_release_tag}")
        self.P(f"Cached asset ID: {self.__last_asset_id}")
        self.P(f"Cached asset updated_at: {self.__last_asset_updated_at}")
      # endif latest_info

      # Auto-extract if needed
      final_extract_dir = download_dir
      if local_file.endswith(".zip"):
        new_dir = self.os_path.join('downloaded_assets', self.plugin_id, 'unzipped_asset')
        self.maybe_download(
          url=local_file,
          fn=new_dir,
          target='output',
          unzip=True
        )
        os.remove(local_file)
        final_extract_dir = self.os_path.join(self.get_output_folder(), new_dir)
      else:
        self.P(f"Prebuilt artifact is not .zip. Leaving as is: {local_file}")

      self.P(f"Prebuilt artifact is ready at {final_extract_dir}")
      return final_extract_dir
    elif operation == "clone":
      username = dct_data.get("username", None)
      token = dct_data.get("token", None)
      track_mode = dct_data.get("track", "commit")  # default = commit
      release_tag_substring = dct_data.get("release_tag", None)

      # Clone the repo into relative_assets_path
      self.git_clone(
        repo_url=assets_path,
        repo_dir=relative_assets_path,
        target='output',
        user=username,
        token=token,
        pull_if_exists=True, # no need to pull if each time we delete the folder
      )

      # If the user wants to track "tag" instead of "commit"
      if track_mode == "tag":
        # 2A) Find the newest release tag (optionally filtered)
        last_release_tag = self.git_get_last_release_tag(
          repo_url=assets_path,
          user=username,
          token=token,
          tag_name=release_tag_substring,  # e.g. "latest" or "mainnet"
        )
        if last_release_tag:
          # 2B) Checkout that release tag in the *already cloned* repo
          full_repo_dir = self.os_path.join(self.get_output_folder(), relative_assets_path)
          self.git_checkout_tag(repo_dir=full_repo_dir, tag=last_release_tag)

          # 2C) Cache the release tag
          self.__last_release_tag = last_release_tag
          self.P(f"Checked out release tag: {last_release_tag}")
        else:
          self.P("No matching release tag found; staying on default HEAD.", color='y')
      # endif release mode

      # 2D) Regardless of commit or release, store the commit hash for debugging
      full_repo_dir = self.os_path.join(self.get_output_folder(), relative_assets_path)
      commit_hash = self.git_get_local_commit_hash(repo_dir=full_repo_dir)
      self.__git_commit_hash = commit_hash
      self.P(f"Finished cloning git repository. Current commit hash: {commit_hash}")
    elif operation == "download":
      self.maybe_download(
        url=assets_path,
        fn=relative_assets_path,
        target='output'
      )
    elif operation == "decode":
      target_dir = self.os_path.join(self.get_output_folder(), relative_assets_path)
      for base64_encoded_file, encoded_file_name in assets_path:
        file_path = self.os_path.join(target_dir, encoded_file_name)
        os.makedirs(self.os_path.dirname(file_path), exist_ok=True)

        encoded_file = self.base64_to_str(base64_encoded_file, decompress=True)
        with open(file_path, 'w') as f:
          f.write(encoded_file)
        # endwith
      # endfor
    else:
      raise ValueError(f"Invalid operation {operation}")

    # now check if it is a zip file
    assets_path = self.os_path.join(self.get_output_folder(), relative_assets_path)
    if self.os_path.isfile(assets_path):
      if not assets_path.endswith('.zip'):
        os.rename(assets_path, assets_path + '.zip')
        assets_path += '.zip'

      relative_assets_path = self.os_path.join('downloaded_assets', self.plugin_id, 'unzipped')
      self.maybe_download(
        url=assets_path,
        fn=relative_assets_path,
        target='output',
        unzip=True
      )
      # remove zip file
      os.remove(assets_path)

    assets_path = self.os_path.join(self.get_output_folder(), relative_assets_path)
    return assets_path

  def __maybe_forced_reload(self):
    reload_interval = self.cfg_forced_reload_interval or 0
    if reload_interval > 0 and self.time() - self.webapp_reload_last_timestamp > reload_interval:
      self.P(f"Forced restart initiated due to `FORCED_RELOAD_INTERVAL` ({reload_interval} seconds)")
      self.__reload_server()
    return

  def __maybe_init_assets(self):
    if self.assets_initialized:
      self.__maybe_forced_reload()
      # check if new git assets are available
      new_repo_version = self.__check_new_repo_version()
      if new_repo_version:
        self.P("New git assets available. Reloading server...")        
        self.__reload_server()
      return
    
    self.P("Initializing assets (`assets_initialized: {}`)".format(self.assets_initialized))

    # download/clone/create/unzip assets
    assets_path = self.__maybe_download_assets()

    # prepare environment variables
    self.__prepare_env(assets_path=assets_path)

    # initialize assets -- copy them to the temp directory
    self.initialize_assets(
      src_dir=assets_path,
      dst_dir=self.script_temp_dir,
      jinja_args=self.jinja_args
    )

    self.webapp_reload_last_timestamp = self.time()
    self.assets_initialized = True
    return

  
  
  # other plugin default methods
  
  def __setup_commands(self):
    self.setup_commands_started = [False] * len(self.get_setup_commands())
    self.setup_commands_finished = [False] * len(self.get_setup_commands())
    self.setup_commands_processes = [None] * len(self.get_setup_commands())
    self.setup_commands_start_time = [None] * len(self.get_setup_commands())

    self.start_commands_started = [False] * len(self.get_start_commands())
    self.start_commands_finished = [False] * len(self.get_start_commands())
    self.start_commands_processes = [None] * len(self.get_start_commands())
    self.start_commands_start_time = [None] * len(self.get_start_commands())

    self.logs = self.deque(maxlen=1000)
    self.dct_logs_reader = {}

    self.err_logs = self.deque(maxlen=1000)
    self.dct_err_logs_reader = {}

    self.P(f"Port: {self.port}")
    self.P(f"Setup commands: {self.get_setup_commands()}")
    self.P(f"Start commands: {self.get_start_commands()}")
    return


  def _on_close(self):
    self.__maybe_close_setup_commands()
    self.__maybe_close_start_commands()

    # close all log readers that are still running
    # this should not have any effect, as the logs should have been
    # closed during the teardown of the setup and start processes
    self.__maybe_read_and_stop_all_log_readers()

    # cleanup the temp directory
    shutil.rmtree(self.script_temp_dir)

    self.__deallocate_port()

    super(BaseWebAppPlugin, self)._on_close()
    return


  def on_close(self):
    # This method is called by super(BaseWebAppPlugin, self)._on_close()
    super(BaseWebAppPlugin, self).on_close()
    self.maybe_stop_ngrok()
    return

  def _on_command(self, data, delta_logs=None, full_logs=None, start=None, reload=None, **kwargs):
    super(BaseWebAppPlugin, self)._on_command(data, **kwargs)

    if (isinstance(data, str) and data.upper() == 'DELTA_LOGS') or delta_logs:
      logs, err_logs = self.__get_delta_logs()
      self.add_payload_by_fields(
        command_params=data,
        logs=logs,
        err_logs=err_logs,
      )
    if (isinstance(data, str) and data.upper() == 'FULL_LOGS') or full_logs:
      # TODO: Implement full logs
      self.add_payload_by_fields(
        command_params=data,
        logs=[]
      )

    if (isinstance(data, str) and data.upper() == 'START') or start:
      self.can_run_start_commands = True
      self.P("Starting server")

    if (isinstance(data, str) and data.upper() == 'RELOAD') or reload:
      self.__reload_server()

    return

  def get_port(self):
    # Override of the auto-generated cfg_port property to handle
    # ports from environment variables.
    configured_port = self.cfg_port
    if isinstance(configured_port, str):
      configured_port = configured_port.strip()
      if configured_port.isdigit():
        configured_port = int(configured_port)
    # endif configured_port is str
    return configured_port

  # Exposed methods
  @property
  def port(self):
    if 'USED_PORTS' not in self.plugins_shmem:
      return None
    if self.str_unique_identification not in self.plugins_shmem['USED_PORTS']:
      return None
    port = self.plugins_shmem['USED_PORTS'][self.str_unique_identification]
    return port

  @property
  def jinja_args(self):
    return {}

  def get_setup_commands(self):
    try:
      return super(BaseWebAppPlugin, self).get_setup_commands() + self.cfg_setup_commands
    except AttributeError:
      return self.cfg_setup_commands

  def get_start_commands(self):
    try:
      return super(BaseWebAppPlugin, self).get_start_commands() + self.cfg_start_commands
    except AttributeError:
      return self.cfg_start_commands

  def initialize_assets(self, src_dir, dst_dir, jinja_args):
    """
    Initialize and copy assets, expanding any jinja templates.
    All files from the source directory are copied to the
    destination directory with the following exceptions:
      - are symbolic links are ignored
      - files named ending with .jinja are expanded as jinja templates,
        .jinja is removed from the filename and the result copied to
        the destination folder.
    This maintains the directory structure of the source folder.
    In case src_dir is None, only the .env files are written in the destination folder.

    Parameters
    ----------
    src_dir: str or None, path to the source directory
    dst_dir: str, path to the destination directory
    jinja_args: dict, jinja keys to use while expanding the templates

    Returns
    -------
    None
    """

    # now copy the assets to the destination
    self.P(f'Copying assets from {src_dir} to {dst_dir} with keys {jinja_args}')

    if src_dir is not None:
      env = Environment(loader=FileSystemLoader('.'))
      # Walk through the source directory.
      for root, _, files in os.walk(src_dir):
        for file in files:
          src_file_path = self.os_path.join(root, file)
          dst_file_path = self.os_path.join(
            dst_dir,
            self.os_path.relpath(src_file_path, src_dir)
          )

          # If we have a symlink don't do anything.
          if self.os_path.islink(src_file_path):
            continue

          # Make sure the destination directory exists.
          os.makedirs(self.os_path.dirname(dst_file_path), exist_ok=True)

          is_jinja_template = False
          if src_file_path.endswith(('.jinja')):
            dst_file_path = dst_file_path[:-len('.jinja')]
            is_jinja_template = True
          if src_file_path.endswith(('.j2')):
            dst_file_path = dst_file_path[:-len('.j2')]
            is_jinja_template = True

          # If this file is a jinja template render it to a file with the .jinja suffix removed.
          # Otherwise just copy the file to the destination directory.
          if is_jinja_template:
            template = env.get_template(src_file_path)
            rendered_content = template.render(jinja_args)
            with open(dst_file_path, 'w') as f:
              f.write(rendered_content)
          else:
            shutil.copy2(src_file_path, dst_file_path)
          # endif is jinja template
        # endfor all files
      # endfor os.walk
    # endif src_dir is not None

    # write .env file in the target directory
    # environment variables are passed in subprocess.Popen, so this is not needed
    # but it's useful for debugging
    with open(self.os_path.join(self.script_temp_dir, '.env'), 'w') as f:
      for key, value in self.prepared_env.items():
        f.write(f"{key}={value}\n")

    with open(self.os_path.join(self.script_temp_dir, '.start_env_used'), 'w') as f:
      for key, value in self.prepared_env.items():
        f.write(f"{key}={value}\n")

    with open(self.os_path.join(self.script_temp_dir, '.setup_env_used'), 'w') as f:
      for key, value in self.base_env.items():
        f.write(f"{key}={value}\n")

    # now cleanup the download folder if it exists
    downloaded_assets_path = self.os_path.join(self.get_output_folder(), 'downloaded_assets', self.plugin_id)
    if self.os_path.exists(downloaded_assets_path):
      shutil.rmtree(downloaded_assets_path)

    self.P("Assets copied successfully")

    return

  def __maybe_ngrok_ping(self):
    # Check if the Ngrok API is used.
    if not self.cfg_ngrok_use_api:
      return
    # Check if the listener is available.
    if self.ngrok_listener is None:
      return
    # Check if the listener has a URL.
    # In case a Ngrok edge label or domain is provided no URL will be available since the user should already have it.
    if self.ngrok_listener.url() is None:
      return
    if self.__last_ngrok_url_ping_ts is None or self.time() - self.__last_ngrok_url_ping_ts >= self.cfg_ngrok_url_ping_interval:
      self.__last_ngrok_url_ping_ts = self.time()
      self.add_payload_by_fields(
        ngrok_url=self.ngrok_listener.url(),
      )
    # endif last ngrok url ping
    return

  def __maybe_setup_commands(self):
    if self.__commands_ready:
      return
    self.__setup_commands()
    self.__commands_ready = True
    return

  def _process(self):  # Check: _process as opposed to process
    self.__maybe_setup_commands()

    self.__maybe_init_assets()  # Check: auto-updates point here?

    self.maybe_init_ngrok()

    self.__maybe_run_all_setup_commands()

    self.maybe_start_ngrok()

    self.__maybe_run_all_start_commands()

    self.__maybe_print_all_logs() # Check: must review 

    super(BaseWebAppPlugin, self)._process() # Check: why is this required
    self.__maybe_ngrok_ping()
    return



try:
  import cv2
except:
  pass
import os
import zipfile
from time import time


def save_images_and_payload_to_output(log, 
                                      np_witness_img, 
                                      np_orig_img, 
                                      path=None,
                                      relative_path=None,
                                      dct_payload=None, 
                                      file_system_manager=None,
                                      last_archive_time=None,
                                      archive_each_minutes=60,
                                      upload_nr_imgs=None, 
                                      perform_upload=False
                                      ):
  if path is not None:
    save_path = path
  elif relative_path is not None:
    save_path = os.path.join(log.get_target_folder('output'), relative_path)
  else:
    raise ValueError('Must specify `path` or `relative_path`')

  dt_now = log.now_str()

  path_witness = os.path.join(save_path, '{}_witness.jpg'.format(dt_now))
  path_img = os.path.join(save_path, '{}_img.jpg'.format(dt_now))
  path_dct = os.path.join(save_path, '{}_dct.json'.format(dt_now))
  path_zip = os.path.join(save_path, f'{dt_now}{"_" + relative_path if relative_path is not None else ""}_archive.zip')

  if np_witness_img is not None:
    if isinstance(np_witness_img, list):
      for i, np_witness_img_shard in enumerate(np_witness_img):
        if np_witness_img_shard is not None:
          path_witness = os.path.join(save_path, '{}_witness_{}.jpg'.format(dt_now, i))
          cv2.imwrite(
            filename=path_witness,
            img=np_witness_img_shard[:, :, ::-1]  # flip channels to BGR
          )
      # endfor
    else:
      cv2.imwrite(
        filename=path_witness,
        img=np_witness_img[:, :, ::-1]  # flip channels to BGR
      )
  # endif

  if np_orig_img is not None:
    cv2.imwrite(
      filename=path_img,
      img=np_orig_img[:, :, ::-1] ### flip channels to BGR
    )
  #endif

  if dct_payload is not None:
    log.save_json(dct_payload, path_dct)
  
  archive_time = None
  
  # upload either if time elapsed or if nr of images is over upload_nr_imgs
  
  # get files
  _lst_files = os.listdir(save_path)
  _lst_files = [
    x for x in _lst_files 
    if ('_witness.jpg' in x) or ('_img.jpg' in x) or ('_dct.json' in x)
  ]
  
  _lst_files = [os.path.join(save_path, x) for x in _lst_files]
  
  n_imgs = len([x for x in _lst_files if '_witness.jpg' in x])

  if (
      (last_archive_time is not None and (time() - last_archive_time) >= (archive_each_minutes * 60)) 
      or 
      (upload_nr_imgs is not None and n_imgs >= upload_nr_imgs)
     ):
    archive_time = time()
    with zipfile.ZipFile(path_zip, 'w') as zipObj:
      for p in _lst_files:
        zipObj.write(p)

    for p in _lst_files:
      # remove all *_witness.jpg, *_img.jpg, *_dct.json files
      os.remove(p)


    url = None
    if perform_upload and file_system_manager is not None:
      url = file_system_manager.upload(
        file_path=path_zip,
        target_path='/DEBUG_SAVE/{}'.format(os.path.basename(path_zip))
      )
    #endif

    if isinstance(dct_payload, dict):
      if url is not None :
        dct_payload['_DBGS_URL'] = url
      else:
        dct_payload['_DBGS_URL'] = path_zip
    #endif

  #endif it is time to archive and maybe upload
  return archive_time


import numpy as np
import os
import torch as th
import torchvision as tv

from naeural_core import Logger



if __name__ == '__main__':

  l = Logger('THTST', base_folder='.', app_folder='_cache', TF_KERAS=False)
  
  train = tv.datasets.MNIST(root=l.get_data_folder(), train=True, download=True)
  test = tv.datasets.MNIST(root=l.get_data_folder(), train=False, download=True)
  
  
  (x_train, y_train), (x_dev, y_dev) = (train.data.numpy(), train.targets.numpy()), (test.data.numpy(), test.targets.numpy())
  x_train = (x_train.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  x_dev = (x_dev.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  y_train = y_train.astype('int64')
  y_dev = y_dev.astype('int64')
  
  l.gpu_info(True)
  
  
  
  fn = os.path.join(l.get_models_folder(), 'simple_model_ts.pt')
  m = th.jit.load(fn)

  dev = next(m.parameters()).device
  
  with th.no_grad():
    for _ in range(100):
      _slice = np.random.randint(0, y_dev.shape[0])
      x_dev_1 = x_dev[:_slice]
      x_dev_2 = x_dev[_slice:]
      th_xd1 = th.tensor(x_dev_1, device=dev)
      th_xd2 = th.tensor(x_dev_2, device=dev)  
      
      l.start_timer('ts_run')
      th_yp1 = m(th_xd1)
      th_yp2 = m(th_xd2)
      l.stop_timer('ts_run')
      
    np_y1 = th_yp1.cpu().numpy().argmax(-1).ravel()
    np_y2 = th_yp2.cpu().numpy().argmax(-1).ravel()
    np_y = np.concatenate((np_y1, np_y2))
  
  res = (np_y == y_dev).sum() / y_dev.shape[0]
  l.P("Dev acc: {:.4f} / Model & data on {}".format(res, dev), color='g')
  l.show_timers()
  
  
  


from ultralytics import YOLO
import torch as th
import json
import os
from pathlib import Path
import shutil
import gc
import torchvision

from naeural_core import Logger

def yolo_models(log):
  model_cfgs = {
     'n': {
      'yaml': 'yolov8n.yaml',
      'weights': 'yolov8n.pt',
    },
    's': {
      'yaml': 'yolov8s.yaml',
      'weights': 'yolov8s.pt',
    },
    'l': {
      'yaml': 'yolov8l.yaml',
      'weights': 'yolov8l.pt',
    },
  }

  sizes = {
    'n': [
      [448, 640],
    ],
    's': [
      [640, 1152],
    ],
    'l': [
      [640, 896],
    ],
    'x': [
      [1152, 2048]
    ]
  }
  classnames_model_path = os.path.join(log.get_models_folder(), '20230723_y8l_nms.ths')
  extra_files = {'config.txt': ''}
  model = th.jit.load(classnames_model_path, map_location='cpu', _extra_files=extra_files)
  CLASSNAMES = json.loads(extra_files['config.txt'])['names']
  model = None
  gc.collect()

  for model_name in sizes.keys():
    gc.collect()
    th.cuda.empty_cache()

    model = YOLO(model_cfgs[model_name]['yaml']).load(model_cfgs[model_name]['weights'])
    device = 'cuda:0'
    model = model.to(device)
    format = "torchscript"

    for imgsz in sizes[model_name]:
      print(f'Exporting y8{model_name} with imgsz={imgsz}')
      export_kwargs = {
        'format': format,
        'imgsz': imgsz,
      }
      pt_path = f'y8{model_name}_{imgsz[0]}x{imgsz[1]}.torchscript'
      setattr(model.model, 'pt_path', pt_path)
      setattr(model.model, 'names', CLASSNAMES)
      yield (model, export_kwargs)

    # endfor imgsz
    # Delete the model and make sure we've freed the memory.
  # endfor model_name

if __name__ == '__main__':
  log = Logger('Y8', base_folder='.', app_folder='_local_cache')
  models_folder = log.get_models_folder()
  for model, export_kwargs in yolo_models(log=log):
    exported = model.export(**export_kwargs)
    shutil.move(model.model.pt_path, os.path.join(models_folder, model.model.pt_path))
    del model
    gc.collect()




class _DeAPIMixin(object):
  def __init__(self):
    self._commands = []
    super(_DeAPIMixin, self).__init__()
    return

  def deapi_get_wokers(self, n_workers):
    return self.netmon.network_top_n_avail_nodes(n_workers)
